from histpy import Histogram, Axis, Axes
import h5py as h5
import sys
from cosipy import SpacecraftFile
from cosipy.response import PointSourceResponse
import healpy as hp
from mhealpy import HealpixMap
import numpy as np
import os
import multiprocessing
from itertools import product
from .fast_norm_fit import FastNormFit as fnf
from pathlib import Path
from cosipy.response import FullDetectorResponse
import time
import scipy.stats
import os
import psutil
import gc
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)

class FastTSMap():
    
    def __init__(self, data, bkg_model, response_path, orientation = None, cds_frame = "local", scheme = "RING"):
        
        """
        Initialize the instance if a TS map fit.
        
        Parameters
        ----------
        data : histpy.Histogram
            Observed data, which includes counts from both signal and background.
        bkg_model : histpy.Histogram
            Background model, which includes the background counts to model the background in the observed data.
        response_path : str or pathlib.Path
            The path to the response file.
        orientation : cosipy.SpacecraftFile, optional
            The orientation of the spacecraft when data are collected (the default is `None`, which implies the orientation file is not needed).
        cds_frame : str, optional
            "local" or "galactic", it's the Compton data space (CDS) frame of the data, bkg_model and the response. In other words, they should have the same cds frame (the default is "local", which implied that a local frame that attached to the spacecraft).
        scheme : str, optional
            The scheme of the CDS of data (the default is "RING", which implies a "RING" scheme of the data).
        """
        
        self._data = data.project(["Em", "PsiChi", "Phi"])
        self._bkg_model = bkg_model.project(["Em", "PsiChi", "Phi"])
        self._orientation = orientation
        self._response_path = Path(response_path)
        self._cds_frame = cds_frame
        self._scheme = scheme
        
    @staticmethod
    def slice_energy_channel(hist, channel_start, channel_stop):
        """
        Slice one or more bins along first axis of the `histogram`.
        
        Parameters
        ----------
        hist : histpy.Histogram
            The histogram object to be sliced.
        channel_start : int
            The start of the slice (inclusive).
        channel_stop : int
            The stop of the slice (exclusive).
        
        Returns
        -------
        sliced_hist : histpy.Histogram
            The sliced histogram.
        """
        
        sliced_hist = hist.slice[channel_start:channel_stop,:]
        
        return sliced_hist
    
    @staticmethod
    def get_hypothesis_coords(nside, scheme = "RING", coordsys = "galactic"):
        
        """
        Get a list of hypothesis coordinates.
        
        Parameters
        ----------
        nside : int
            The nside of the map.
        scheme : str, optional
            The scheme of the map where the hypothesis coordinates are generated (the default is "RING", which implies the "RING" scheme is used to get the hypothesis coordinates).
        coordsys : str, optional
            The coordinate system used in the map where the hypothesis coordinates are generated (the default is "galactic", which implies the galactic coordinates system is used).
        
        Returns
        -------
        hypothesis_coords : list
            The list of the hypothesis coordinates at the center of each pixel.
        """
        
        data_array = np.zeros(hp.nside2npix(nside))
        ts_temp = HealpixMap(data = data_array, scheme = scheme, coordsys = coordsys)
        
        hypothesis_coords = []
        for i in np.arange(data_array.shape[0]):
            hypothesis_coords += [ts_temp.pix2skycoord(i)]
            
        return hypothesis_coords
    
    
    @staticmethod
    def get_cds_array(hist, energy_channel):
        
        """
        Get the flattened cds array from input Histogram.
        
        Parameters
        -----------
        hist : histpy.Histogram
            The input Histogram.
        energy_channel : list
            The format is `[lower_channel, upper_chanel]`. The lower_channel is inclusive while the upper_channel is exclusive.

        Returns
        -------
        cds_array : numpy.ndarray
            The flattended Compton data space (CDS) array.
        
        """
        if not isinstance(hist, Histogram):
            raise TypeError("Please input hist must be a histpy.Histogram object.")
        
        hist_axes_labels = hist.axes.labels
        cds_labels = ["PsiChi", "Phi"]
        if not all([label in hist_axes_labels for label in cds_labels]):
            raise ValueError("The data doesn't contain the full Compton Data Space!")
            
        hist = hist.project(["Em", "PsiChi", "Phi"]) # make sure the first axis is the measured energy
        hist_cds_sliced = FastTSMap.slice_energy_channel(hist, energy_channel[0], energy_channel[1])   
        hist_cds = hist_cds_sliced.project(["PsiChi", "Phi"])
        cds_array = np.array(hist_cds.to_dense()[:]).flatten()  # here [:] is equivalent to [:, :]
        del hist
        del hist_cds_sliced
        del hist_cds
        gc.collect()
        
        return cds_array
    
    @staticmethod
    def get_psr_in_galactic(hypothesis_coord, response_path, spectrum):
        
        """
        Get the point source response (psr) in galactic. Please be aware that you must use a galactic response!
        To do: to make the weight parameter not hardcoded
        
        Parameters
        ----------
        hypothesis_coord : astropy.coordinates.SkyCoord
            The hypothesis coordinate.
        response_path : str or path.lib.Path
            The path to the response.
        spectrum : astromodels.functions
            The spectrum of the source to be placed at the hypothesis coordinate.
        
        Returns
        -------
        psr : histpy.Histogram
            The point source response of the spectrum at the hypothesis coordinate.
        """
        
        # Open the response
        # Notes from Israel: Inside it contains a single histogram with all the regular axes for a Compton Data Space (CDS) analysis, in galactic coordinates. Since there is no class yet to handle it, this is how to read in the HDF5 manually.
        
        with h5.File(response_path) as f:

            axes_group = f['hist/axes']
            axes = []
            for axis in axes_group.values():
                # Get class. Backwards compatible with version
                # with only Axis
                axis_cls = Axis
                if '__class__' in axis.attrs:
                    class_module, class_name = axis.attrs['__class__']
                    axis_cls = getattr(sys.modules[class_module], class_name)
                axes += [axis_cls._open(axis)]
        axes = Axes(axes)
        
        # get the pixel number of the hypothesis coordinate
        map_temp = HealpixMap(base = axes[0])
        hypothesis_coord_pix_number = map_temp.ang2pix(hypothesis_coord)
        
        # get the expectation for the hypothesis coordinate (a point source)
        with h5.File(response_path) as f:
            pix = hypothesis_coord_pix_number
            psr = PointSourceResponse(axes[1:], f['hist/contents'][pix+1], unit = f['hist'].attrs['unit'])
                
        return psr
    
    
    @staticmethod
    def get_ei_cds_array(hypothesis_coord, energy_channel, response_path, spectrum, cds_frame, orientation = None):
                         
        """
        Get the expected counts in CDS in local or galactic frame.
        
        Parameters
        ----------
        hypothesis_coord : astropy.coordinates.SkyCoord
            The hypothesis coordinate. 
        energy_channel : list
            The format is `[lower_channel, upper_chanel]`. The lower_channel is inclusive while the upper_channel is exclusive.
        response_path : str or pathlib.Path
            The path to the response file.
        spectrum : astromodels.functions
            The spectrum of the source.
        cds_frame : str, optional
            "local" or "galactic", it's the Compton data space (CDS) frame of the data, bkg_model and the response. In other words, they should have the same cds frame.
        orientation : cosipy.spacecraftfile.SpacecraftFile, optional
            The orientation of the spacecraft when data are collected (the default is `None`, which implies the orientation file is not needed).
        
        Returns
        -------
        cds_array : numpy.ndarray
            The flattended Compton data space (CDS) array.
        """
                         
        # check inputs, will complete later
        
        # the local and galactic frame works very differently, so we need to compuate the point source response (psr) accordingly
        #time_cds_start = time.time()
        if cds_frame == "local":
            
            if orientation == None:
                raise TypeError("The when the data are binned in local frame, orientation must be provided to compute the expected counts.")

            #time_coord_convert_start = time.time()
            # convert the hypothesis coord to the local frame (Spacecraft frame)
            hypothesis_in_sc_frame = orientation.get_target_in_sc_frame(target_name = "Hypothesis", 
                                                                        target_coord = hypothesis_coord, 
                                                                        quiet = True)
            #time_coord_convert_end = time.time()
            #time_coord_convert_used = time_coord_convert_end - time_coord_convert_start
            #logger.info(f"The time used for coordinate conversion is {time_coord_convert_used}s.")

            #time_dwell_start = time.time()
            # get the dwell time map: the map of the time spent on each pixel in the local frame
            dwell_time_map = orientation.get_dwell_map(response = response_path)
            #time_dwell_end = time.time()
            #time_dwell_used = time_dwell_end - time_dwell_start
            #logger.info(f"The time used for dwell time map is {time_dwell_used}s.")
            
            #time_psr_start = time.time()
            # convolve the response with the dwell_time_map to get the point source response
            with FullDetectorResponse.open(response_path) as response:
                psr = response.get_point_source_response(dwell_time_map)
            #time_psr_end = time.time()
            #time_psr_used = time_psr_end - time_psr_start
            #logger.info(f"The time used for psr is {time_psr_used}s.")

        elif cds_frame == "galactic":
            
            psr = FastTSMap.get_psr_in_galactic(hypothesis_coord = hypothesis_coord, response_path = response_path, spectrum = spectrum)
            
        else:
            raise ValueError("The point source response must be calculated in the local and galactic frame. Others are not supported (yet)!")
            
        # convolve the point source reponse with the spectrum to get the expected counts
        expectation = psr.get_expectation(spectrum)
        del psr
        gc.collect()

        # slice energy channals and project it to CDS
        ei_cds_array = FastTSMap.get_cds_array(expectation, energy_channel)
        del expectation
        gc.collect()

        #time_cds_end = time.time()
        #time_cds_used = time_cds_end - time_cds_start
        #logger.info(f"The time used for cds is {time_cds_used}s.")
        
        return ei_cds_array
    
    @staticmethod
    def fast_ts_fit(hypothesis_coord, 
                    energy_channel, data_cds_array, bkg_model_cds_array, 
                    orientation, response_path, spectrum, cds_frame,
                    ts_nside, ts_scheme):

        """
        Perform a TS fit on a single location at `hypothesis_coord`.

        Parameters
        ----------
        hypothesis_coord : astropy.coordinates.SkyCoord
            The hypothesis coordinate. 
        energy_channel : list
            The format is `[lower_channel, upper_chanel]`. The lower_channel is inclusive while the upper_channel is exclusive.
        data_cds_array : numpy.ndarray
            The flattened Compton data space (CDS) array of the data.
        bkg_model_cds_array : numpy.ndarray
            The flattened Compton data space (CDS) array of the background model.
        orientation : cosipy.spacecraftfile.SpacecraftFile
            The orientation of the spacecraft when data are collected. 
        response_path : str or pathlib.Path
            The path to the response file.
        spectrum : astromodels.functions
            The spectrum of the source.
        cds_frame : str
            "local" or "galactic", it's the Compton data space (CDS) frame of the data, bkg_model and the response. In other words, they should have the same cds frame .
        ts_nside : int
            The nside of the ts map.   
        ts_scheme : str
            The scheme of the Ts map.

        Returns
        -------
        list
            The list of the resulting TS fit: [pix number, ts value, norm, norm_err, failed, iterations, time_ei_cds_array, time_fit, time_fast_ts_fit]
        """
        
        start_fast_ts_fit = time.time()

        # get the indices of the pixels to fit
        pix = hp.ang2pix(nside = ts_nside, theta = hypothesis_coord.l.deg, phi = hypothesis_coord.b.deg, lonlat = True)
        
        # get the expected counts in the flattened cds array
        start_ei_cds_array = time.time()
        ei_cds_array = FastTSMap.get_ei_cds_array(hypothesis_coord = hypothesis_coord, cds_frame = cds_frame,
                                                  energy_channel = energy_channel, orientation = orientation, 
                                                  response_path = response_path, spectrum = spectrum)
        end_ei_cds_array = time.time()
        time_ei_cds_array = end_ei_cds_array - start_ei_cds_array
        
        # start the fit
        start_fit = time.time()
        fit = fnf(max_iter=1000)
        result = fit.solve(data_cds_array, bkg_model_cds_array, ei_cds_array)
        end_fit = time.time()
        time_fit = end_fit - start_fit

        end_fast_ts_fit = time.time()
        time_fast_ts_fit = end_fast_ts_fit - start_fast_ts_fit
        
        return [pix, result[0], result[1], result[2], result[3], result[4], time_ei_cds_array, time_fit, time_fast_ts_fit]

        
    def parallel_ts_fit(self, hypothesis_coords, energy_channel, spectrum, ts_scheme = "RING", start_method = "fork", cpu_cores = None, ts_nside = None):
        
        """
        Perform parallel computation on all the hypothesis coordinates.
        
        Parameters
        ----------
        hypothesis_coords : list
            A list of the hypothesis coordinates to fit
        energy_channel : list
            the energy channel you want to use: [lower_channel, upper_channel]. lower_channel is inclusive while upper_channel is exclusive.
        spectrum : astromodels.functions
            The spectrum of the source.
        ts_scheme : str, optional
            The scheme of the TS map (the default is "RING", which implies a "RING" scheme of the TS map).
        start_method : str, optional
            The starting method of the parallel computation (the default is "fork", which implies using the fork method to start parallel computation).
        cpu_cores : int, optional
            The number of cpu cores you wish to use for the parallel computation (the default is `None`, which implies using all the available number of cores -1 to perform the parallel computation).
        ts_nside : int, optional
            The nside of the ts map. This must be given if the number of hypothesis_coords isn't equal to the number of pixels of the total ts map, which means that you fit only a portion of the total ts map. (the default is `None`, which means that you fit the full ts map). 
        
        Returns
        -------
        results : numpy.ndarray
            The result of the ts fit over all the hypothesis coordinates.
        """
        
        # decide the ts_nside from the list of hypothesis coordinates if not given
        if ts_nside == None:
            ts_nside = hp.npix2nside(len(hypothesis_coords))
        
        # get the flattened data_cds_array
        data_cds_array = FastTSMap.get_cds_array(self._data, energy_channel).flatten()
        bkg_model_cds_array = FastTSMap.get_cds_array(self._bkg_model, energy_channel).flatten()
        
        if (data_cds_array[bkg_model_cds_array ==0]!=0).sum() != 0:
            #raise ValueError("You have data!=0 but bkg=0, check your inputs!")
            # let's try to set the data bin to zero when the corresponding bkg bin isn't zero.
            # Need further investigate, why bkg = 0 but data!=0 happens? ==> it's more like an issue related to simulated data instead of code
            # This first happened in GRB fitting, but got fixed somehow <== I now understand it's caused by using different PsiChi binning in the same fit
            # But it also happened to Crab while the PsiChi binning are both galactic for Crab and the Albedo, why???? ?_?
            data_cds_array[bkg_model_cds_array == 0] =0
            
        
        # set up the number of cores to use for the parallel computation
        total_cores = multiprocessing.cpu_count()
        if cpu_cores == None or cpu_cores >= total_cores:
            # if you don't specify the number of cpu cores to use or the specified number of cpu cores is the same as the total number of cores you have
            # it will use the [total_cores - 1] number of cores to run the parallel computation.
            cores = total_cores - 1
            logger.info(f"You have total {total_cores} CPU cores, using {cores} CPU cores for parallel computation.")
        else:
            cores = cpu_cores
            logger.info(f"You have total {total_cores} CPU cores, using {cores} CPU cores for parallel computation.")

        start = time.time()
        multiprocessing.set_start_method(start_method, force = True)
        pool = multiprocessing.Pool(processes = cores)
        results = pool.starmap(FastTSMap.fast_ts_fit, product(hypothesis_coords, [energy_channel], [data_cds_array], [bkg_model_cds_array], 
                                                             [self._orientation], [self._response_path], [spectrum], [self._cds_frame], 
                                                             [ts_nside], [ts_scheme]))
            
        pool.close()
        pool.join()
        
        end = time.time()
        
        elapsed_seconds = end - start
        elapsed_minutes = elapsed_seconds/60
        logger.info(f"The time used for the parallel TS map computation is {elapsed_minutes} minutes")
        
        results = np.array(results)  # turn to a numpy array
        results = results[results[:, 0].argsort()]  # arrange the order by the pixel numbering
        self.result_array = results  # the full result array
        self.ts_array = results[:,1]  # the ts array
        
        return results

    @staticmethod
    def _plot_ts(ts_array, skycoord = None, containment = None, save_plot = False, save_dir = "", save_name = "ts_map.png", dpi = 300):

        """
        Plot the containment region of the TS map.

        Parameters
        ----------
        ts_array : numpy.ndarray
            The array of ts values from parallel ts fit.
        skyoord : astropy.coordinates.SkyCoord, optional
            The true location of the source (the default is `None`, which implies that there are no coordiantes to be printed on the TS map).
        containment : float, optional
            The containment level of the source (the default is `None`, which will plot raw TS values).
        save_plot : bool, optional
            Set `True` to save the plot (the default is `False`, which means it won't save the plot.
        save_dir : str or pathlib.Path, optional
            The directory to save the plot.
        save_name : str, optional
            The file name of the plot to be save.
        dpi : int, optional
            The dpi for plotting and saving.
        """


        if skycoord != None:
            lon = skycoord.l.deg
            lat = skycoord.b.deg


        # get the ts value
        m_ts = ts_array
        
        # get plotting canvas
        fig, ax = plt.subplots(dpi=dpi)
        
        # plot the ts map with containment region
        if containment != None:
            critical = FastTSMap.get_chi_critical_value(containment = containment)
            percentage = containment*100
            max_ts = np.max(m_ts[:])
            min_ts = np.min(m_ts[:])        
            hp.mollview(m_ts[:], max = max_ts, min = max_ts-critical, title = f"Containment {percentage}%", coord = "G", hold = True) 
        elif containment == None:
            hp.mollview(m_ts[:], coord = "G", hold = True) 

        if skycoord != None:
            hp.projscatter(lon, lat, marker = "x", linewidths = 0.5, lonlat=True, coord = "G", label = f"True location at l={lon}, b={lat}", color = "fuchsia")
        hp.projscatter(0, 0, marker = "o", linewidths = 0.5, lonlat=True, coord = "G", color = "red")
        hp.projtext(350, 0, "(l=0, b=0)", lonlat=True, coord = "G", color = "red")

        if save_plot == True:

            fig.savefig(Path(save_dir)/save_name, dpi = dpi)

        return

    def plot_ts(self, ts_array = None, skycoord = None, containment = None, save_plot = False, save_dir = "", save_name = "ts_map.png", dpi = 300):

        """
        Plot the containment region of the TS map.

        Parameters
        ----------
        skyoord : astropy.coordinates.SkyCoord, optional
            The true location of the source (the default is `None`, which implies that there are no coordiantes to be printed on the TS map).
        containment : float, optional
            The containment level of the source (the default is `None`, which will plot raw TS values).
        save_plot : bool, optional
            Set `True` to save the plot (the default is `False`, which means it won't save the plot.
        save_dir : str or pathlib.Path, optional
            The directory to save the plot.
        save_name : str, optional
            The file name of the plot to be save.
        dpi : int, optional
            The dpi for plotting and saving.
        """

        if ts_array is not None:

            FastTSMap._plot_ts(ts_array = ts_array, skycoord = skycoord, containment = containment, 

                               save_plot = save_plot, save_dir = save_dir, save_name = save_name, dpi = dpi)

        else:

            FastTSMap._plot_ts(ts_array = self.ts_array, skycoord = skycoord, containment = containment, 
                               save_plot = save_plot, save_dir = save_dir, save_name = save_name, dpi = dpi)

        return

    @staticmethod
    def get_chi_critical_value(containment = 0.90):
        
        """
        Get the critical value of the chi^2 distribution based ob the confidence level.

        Parameters
        ----------
        containment : float, optional
            The confidence level of the chi^2 distribution (the default is `0.9`, which implies that the 90% containment region).

        Returns
        -------
        float
            The critical value corresponds to the confidence level.
        """

        return scipy.stats.chi2.ppf(containment, df=2)

    @staticmethod
    def show_memory_info(hint):
        pid = os.getpid()
        p = psutil.Process(pid)
    
        info = p.memory_full_info()
        memory = info.uss / 1024. / 1024
        logger.info('{} memory used: {} MB'.format(hint, memory))
        
    


from threeML import PluginPrototype
from threeML.minimizer import minimization
from threeML.config.config import threeML_config
from threeML.exceptions.custom_exceptions import FitFailed
from astromodels import Parameter

from cosipy.response.FullDetectorResponse import FullDetectorResponse
from cosipy.response.ExtendedSourceResponse import ExtendedSourceResponse
from cosipy.image_deconvolution import AllSkyImageModel

from scoords import SpacecraftFrame, Attitude

from mhealpy import HealpixMap

from cosipy.response import PointSourceResponse, DetectorResponse
from histpy import Histogram
import h5py as h5
from histpy import Axis, Axes
import sys

import astropy.units as u
import astropy.coordinates as coords

from sparse import COO

import numpy as np

from scipy.special import factorial

import collections

import copy

import logging
logger = logging.getLogger(__name__)

import inspect

class COSILike(PluginPrototype):
    """
    COSI 3ML plugin.

    Parameters
    ----------
    name : str
        Plugin name e.g. "cosi". Needs to have a distinct name with respect to other plugins in the same analysis
    dr : str
        Path to full detector response
    data : histpy.Histogram
        Binned data. Note: Eventually this should be a cosipy data class
    bkg : histpy.Histogram
        Binned background model. Note: Eventually this should be a cosipy data class
    sc_orientation : cosipy.spacecraftfile.SpacecraftFile
        Contains the information of the orientation: timestamps (astropy.Time) and attitudes (scoord.Attitude) that describe
        the spacecraft for the duration of the data included in the analysis
    nuisance_param : astromodels.core.parameter.Parameter, optional
        Background parameter
    coordsys : str, optional
        Coordinate system ('galactic' or 'spacecraftframe') to perform fit in, which should match coordinate system of data 
        and background. This only needs to be specified if the binned data and background do not have a coordinate system 
        attached to them
    precomputed_psr_file : str, optional
        Full path to precomputed point source response in Galactic coordinates
    earth_occ : bool, optional
        Option to include Earth occultation in fit (default is True).
    """
    def __init__(self, name, dr, data, bkg, sc_orientation, 
                 nuisance_param=None, coordsys=None, precomputed_psr_file=None, earth_occ=True, **kwargs):
        
        # create the hash for the nuisance parameters. We have none for now.
        self._nuisance_parameters = collections.OrderedDict()

        # call the prototype constructor. Boilerplate.
        super(COSILike, self).__init__(name, self._nuisance_parameters)

        # User inputs needed to compute the likelihood
        self._name = name
        self._rsp_path = dr
        self._dr = FullDetectorResponse.open(dr)
        self._data = data
        self._bkg = bkg
        self._sc_orientation = sc_orientation
        self.earth_occ = earth_occ
        
        try:
            if data.axes["PsiChi"].coordsys.name != bkg.axes["PsiChi"].coordsys.name:
                raise RuntimeError("Data is binned in " + data.axes["PsiChi"].coordsys.name + " and background is binned in " 
                                   + bkg.axes["PsiChi"].coordsys.name + ". They should be binned in the same coordinate system.")
            else:
                self._coordsys = data.axes["PsiChi"].coordsys.name
        except:
            if coordsys == None:
                raise RuntimeError(f"There is no coordinate system attached to the binned data. One must be provided by " 
                                   f"specifiying coordsys='galactic' or 'spacecraftframe'")
            else:
                self._coordsys = coordsys
            
        # Place-holder for cached data.
        self._model = None
        self._source = None
        self._psr = None
        self._signal = None
        self._expected_counts = None 

        # Set to fit nuisance parameter if given by user
        if nuisance_param == None:
            self.set_inner_minimization(False)
        elif isinstance(nuisance_param, Parameter):
            self.set_inner_minimization(True)
            self._bkg_par = nuisance_param
            self._nuisance_parameters[self._bkg_par.name] = self._bkg_par
            self._nuisance_parameters[self._bkg_par.name].free = self._fit_nuisance_params
        else:
            raise RuntimeError("Nuisance parameter must be astromodels.core.parameter.Parameter object")
        
        # Option to use precomputed point source response.
        # Note: this still needs to be implemented in a 
        # consistent way for point srcs and extended srcs. 
        self.precomputed_psr_file = precomputed_psr_file
        if self.precomputed_psr_file != None:
            logger.info("... loading the pre-computed image response ...")
            self.image_response = ExtendedSourceResponse.open(self.precomputed_psr_file)
            logger.info("--> done")
        
    def set_model(self, model):
        """
        Set the model to be used in the joint minimization.
        
        Parameters
        ----------
        model : astromodels.core.model.Model
            Any model supported by astromodels
        """
        
        # Temporary fix to only print log-likelihood warning once max per fit
        if inspect.stack()[1][3] == '_assign_model_to_data':
            self._printed_warning = False
    
        # Get point sources and extended sources from model: 
        point_sources = model.point_sources
        extended_sources = model.extended_sources
        
        # Source counter for models with multiple sources:
        self.src_counter = 0
        
        # Get expectation for extended sources:
        
        # Save expected counts for each source,
        # in order to enable easy plotting after likelihood scan:
        if self._expected_counts == None:
            self._expected_counts = {}

        for name,source in extended_sources.items():

            # Set spectrum:
            # Note: the spectral parameters are updated internally by 3ML
            # during the likelihood scan. 

            # Get expectation using precomputed psr in Galactic coordinates:
            total_expectation = self.image_response.get_expectation_from_astromodel(source)

            # Save expected counts for source:
            self._expected_counts[name] = copy.deepcopy(total_expectation)

            # Need to check if self._signal type is dense (i.e. 'Quantity') or sparse (i.e. 'COO').
            if type(total_expectation.contents) == u.quantity.Quantity:
                total_expectation = total_expectation.contents.value
            elif type(total_expectation.contents) == COO:
                total_expectation = total_expectation.contents.todense() 
            else:
                raise RuntimeError("Expectation is an unknown object")

            # Add source to signal and update source counter:
            if self.src_counter == 0:
                self._signal = total_expectation
            if self.src_counter != 0:
                self._signal += total_expectation
            self.src_counter += 1

        # Initialization
        # probably it is better that this part be outside of COSILike (HY).
        if len(point_sources) != 0:
        
            if self._psr is None or len(point_sources) != len(self._psr):

                logger.info("... Calculating point source responses ...")

                self._psr = {}
                self._source_location = {} # Should the poition information be in the point source response? (HY)

                for name, source in point_sources.items():
                    coord = source.position.sky_coord
                
                    self._source_location[name] = copy.deepcopy(coord) # to avoid same memory issue

                    if self._coordsys == 'spacecraftframe':
                        dwell_time_map = self._get_dwell_time_map(coord)
                        self._psr[name] = self._dr.get_point_source_response(exposure_map=dwell_time_map)
                    elif self._coordsys == 'galactic':
                        scatt_map = self._get_scatt_map(coord)
                        self._psr[name] = self._dr.get_point_source_response(coord=coord, scatt_map=scatt_map)
                    else:
                        raise RuntimeError("Unknown coordinate system")

                    logger.info(f"--> done (source name : {name})")

                logger.info(f"--> all done")
        
        # check if the source location is updated or not
        for name, source in point_sources.items():

            if source.position.sky_coord != self._source_location[name]:
                logger.info(f"... Re-calculating the point source response of {name} ...")
                coord = source.position.sky_coord

                self._source_location[name] = copy.deepcopy(coord) # to avoid same memory issue
                
                if self._coordsys == 'spacecraftframe':
                    dwell_time_map = self._get_dwell_time_map(coord)
                    self._psr[name] = self._dr.get_point_source_response(exposure_map=dwell_time_map)
                elif self._coordsys == 'galactic':
                    scatt_map = self._get_scatt_map()
                    self._psr[name] = self._dr.get_point_source_response(coord=coord, scatt_map=scatt_map)
                else:
                    raise RuntimeError("Unknown coordinate system")

                logger.info(f"--> done (source name : {name})")

        # Get expectation for point sources:
        for name,source in point_sources.items():

            # Convolve with spectrum
            # See also the Detector Response and Source Injector tutorials
            spectrum = source.spectrum.main.shape

            total_expectation = self._psr[name].get_expectation(spectrum)
            
            # Save expected counts for source:
            self._expected_counts[name] = copy.deepcopy(total_expectation)
         
            # Need to check if self._signal type is dense (i.e. 'Quantity') or sparse (i.e. 'COO').
            if type(total_expectation.contents) == u.quantity.Quantity:
                total_expectation = total_expectation.project(['Em', 'Phi', 'PsiChi']).contents.value
            elif type(total_expectation.contents) == COO:
                total_expectation = total_expectation.project(['Em', 'Phi', 'PsiChi']).contents.todense() 
            else:
                raise RuntimeError("Expectation is an unknown object")

            # Add source to signal and update source counter:
            if self.src_counter == 0:
                self._signal = total_expectation
            if self.src_counter != 0:
                self._signal += total_expectation
            self.src_counter += 1

        # Cache
        self._model = model

    def get_log_like(self):
        """
        Calculate the log-likelihood.
        
        Returns
        ----------
        log_like : float
            Value of the log-likelihood
        """
        
        # Recompute the expectation if any parameter in the model changed
        if self._model is None:
            log.error("You need to set the model first")
       
        # Set model:
        self.set_model(self._model)
        
        # Compute expectation including free background parameter:
        if self._fit_nuisance_params: 
            if type(self._bkg.contents) == COO:
                expectation = self._signal + self._nuisance_parameters[self._bkg_par.name].value * self._bkg.contents.todense()
            else:
                expectation = self._signal + self._nuisance_parameters[self._bkg_par.name].value * self._bkg.contents
        
        # Compute expectation without background parameter:
        else: 
            if type(self._bkg.contents) == COO:
                expectation = self._signal + self._bkg.contents.todense()
            else:
                expectation = self._signal + self._bkg.contents

        expectation += 1e-12 # to avoid -infinite log-likelihood (occurs when expected counts = 0 but data != 0)
        if not self._printed_warning:
            logger.warning("Adding 1e-12 to each bin of the expectation to avoid log-likelihood = -inf.")
            self._printed_warning = True
        # This 1e-12 should be defined as a parameter in the near future (HY)
        
        # Convert data into an arrary:
        data = self._data.contents
        
        # Compute the log-likelihood:
        log_like = np.nansum(data*np.log(expectation) - expectation)
        
        return log_like
    
    def inner_fit(self):
        """
        Required for 3ML fit.
        """
        
        return self.get_log_like()
    
    def _get_dwell_time_map(self, coord):
        """
        Get the dwell time map of the source in the inertial (spacecraft) frame.
        
        Parameters
        ----------
        coord : astropy.coordinates.SkyCoord
            Coordinates of the target source
        
        Returns
        -------
        dwell_time_map : mhealpy.containers.healpix_map.HealpixMap
            Dwell time map
        """
        
        self._sc_orientation.get_target_in_sc_frame(target_name = self._name, target_coord = coord)
        dwell_time_map = self._sc_orientation.get_dwell_map(response = self._rsp_path)
        
        return dwell_time_map
    
    def _get_scatt_map(self, coord):
        """
        Get the spacecraft attitude map of the source in the inertial (spacecraft) frame.
        
        Parameters
        ----------
        coord : astropy.coordinates.SkyCoord
            The coordinates of the target object.

        Returns
        -------
        scatt_map : cosipy.spacecraftfile.scatt_map.SpacecraftAttitudeMap
        """
        
        scatt_map = self._sc_orientation.get_scatt_map(coord, nside = self._dr.nside * 2, \
                coordsys = 'galactic', earth_occ = self.earth_occ)
        
        return scatt_map
    
    def set_inner_minimization(self, flag: bool):
        """
        Turn on the minimization of the internal COSI (nuisance) parameters.
        
        Parameters
        ----------
        flag : bool
            Turns on and off the minimization  of the internal parameters
        """
        
        self._fit_nuisance_params: bool = bool(flag)

        for parameter in self._nuisance_parameters:
            self._nuisance_parameters[parameter].free = self._fit_nuisance_params



from .PointSourceResponse import PointSourceResponse
from .DetectorResponse import DetectorResponse
from .ExtendedSourceResponse import ExtendedSourceResponse
from astromodels.core.model_parser import ModelParser
import matplotlib.pyplot as plt
from astropy.time import Time
from astropy.coordinates import SkyCoord
from astropy.units import Quantity
import astropy.units as u
from sparse import COO
from pathlib import Path
import numpy as np
import mhealpy as hp
from mhealpy import HealpixBase, HealpixMap
import glob

from scipy.special import erf

from yayc import Configurator

from scoords import SpacecraftFrame, Attitude

from histpy import Histogram, Axes, Axis, HealpixAxis
import h5py as h5
import os
import textwrap
import argparse
import logging
logger = logging.getLogger(__name__)

from copy import copy, deepcopy
import gzip
#from tqdm import tqdm
from tqdm.autonotebook import tqdm
import subprocess
import sys
import pathlib
import gc

class FullDetectorResponse(HealpixBase):
    """
    Handles the multi-dimensional matrix that describes the
    full all-sky response of the instrument.

    You can access the :py:class:`DetectorResponse` at a given pixel using the ``[]``
    operator. Alternatively you can obtain the interpolated reponse using
    :py:func:`get_interp_response`.
    """

    def __init__(self, *args, **kwargs):
        # Overload parent init. Called in class methods.
        pass

    @classmethod
    def open(cls, filename,Spectrumfile=None,norm="Linear" ,single_pixel = False,alpha=0,emin=90,emax=10000, polarization=False):
        """
        Open a detector response file.

        Parameters
        ----------
        filename : str, :py:class:`~pathlib.Path`
        Path to the response file (.h5 or .rsp)

        Spectrumfile : str, 
             path to the input spectrum file used
             for the simulation (optional).

         norm : str, 
             type of normalisation : file (then specify also SpectrumFile)
             ,powerlaw, Mono or Linear
         
         alpha : int,
             if the normalisation is "powerlaw", value of the spectral index.

         single_pixel : bool,
             True if there is only one pixel and not full-sky.

         emin,emax : float
             emin/emax used in the simulation source file.  
        """
        
        filename = Path(filename)


        if filename.suffix == ".h5":
            return cls._open_h5(filename)
        elif "".join(filename.suffixes[-2:]) == ".rsp.gz":
            return cls._open_rsp(filename,Spectrumfile,norm ,single_pixel,alpha,emin,emax)
        else:
            raise ValueError(
                "Unsupported file format. Only .h5 and .rsp.gz extensions are supported.")

    @classmethod
    def _open_h5(cls, filename):
        """
         Open a detector response h5 file.

         Parameters
         ----------
         filename : str, :py:class:`~pathlib.Path`
             Path to HDF5 file
         """
        new = cls(filename)

        new._file = h5.File(filename, mode='r')

        new._drm = new._file['DRM']

        new._unit = u.Unit(new._drm.attrs['UNIT'])
        
        try:
             new._sparse = new._drm.attrs['SPARSE']
        except KeyError:
             new._sparse = True

        # Axes
        axes = []

        for axis_label in new._drm["AXES"]:

            axis = new._drm['AXES'][axis_label]

            axis_type = axis.attrs['TYPE']

            if axis_type == 'healpix':

                axes += [HealpixAxis(edges=np.array(axis),
                                         nside=axis.attrs['NSIDE'],
                                         label=axis_label,
                                         scheme=axis.attrs['SCHEME'],
                                         coordsys=SpacecraftFrame())]

            else:
                axes += [Axis(np.array(axis) * u.Unit(axis.attrs['UNIT']),
                                  scale=axis_type,
                                  label=axis_label)]

        new._axes = Axes(axes)

        # Init HealpixMap (local coordinates, main axis)
        HealpixBase.__init__(new,
                                 base=new.axes['NuLambda'],
                                 coordsys=SpacecraftFrame())

        return new

    @classmethod
    def _open_rsp(cls, filename, Spectrumfile=None,norm="Linear" ,single_pixel = False,alpha=0,emin=90,emax=10000):
        """
        
         Open a detector response rsp file.

         Parameters
         ----------
         filename : str, :py:class:`~pathlib.Path`
             Path to rsp file

         Spectrumfile : str, 
             path to the input spectrum file used
             for the simulation (optional).

         norm : str, 
             type of normalisation : file (then specify also SpectrumFile)
             ,powerlaw, Mono or Linear
         
         alpha : int,
             if the normalisation is "powerlaw", value of the spectral index.

         single_pixel : bool,
             True if there is only one pixel and not full-sky.

         emin,emax : float
             emin/emax used in the simulation source file.
        """

        
        
        axes_names = []
        axes_edges = []
        axes_types = []
        sparse = None
        # get the header infos of the rsp file (nsim,area,bin_edges,etc...)
        with gzip.open(filename, "rt") as file:
            for n, line in enumerate(file):
    
                line = line.split()

                if len(line) == 0:
                    continue

                key = line[0]

                if key == 'TS':
                    nevents_sim = int(line[1])

                elif key == 'SA':
                    area_sim = float(line[1])
                    
                elif key == "SP" :
                    
                    try :
                        norm = str(line[1])
                    except :
                        logger.info(f"norm not found in the file ! We assume {norm}")

                    if norm =="Linear" :
                        emin = int(line[2])
                        emax = int(line[3])
                    
                    if norm == "Gaussian" :
                        Gauss_mean = float(line[2])   
                        Gauss_sig = float(line[3])
                        Gauss_cutoff = float(line[4])          
  	
                elif key == "MS":
                    if line[1] == "true" :
                        sparse = True
                    if line[1] == "false" :
                        sparse = False

                elif key == 'AN':
                    axes_names += [" ".join(line[1:])]

                elif key == 'AD':

                    if axes_types[-1] == "FISBEL":

                        raise RuntimeError("FISBEL binning not currently supported")
                        
                    elif axes_types[-1] == "HEALPix":

                        if line[2] != "RING":
                            raise RuntimeError(f"Scheme {line[2]} not supported")

                        if line[1] == '-1':
                            # Single bin axis --i.e. all-sky
                            axes_edges.append(-1)
                        else:
                            nside = int(2**int(line[1]))
                            axes_edges.append(int(12*nside**2))
                        
                        
                    else:
                        
                        axes_edges.append(np.array(line[1:], dtype='float'))

                elif key == 'AT':
                    axes_types += [line[2]]

                elif key == 'RD':
                    break
                
                elif key == "StartStream":
                    nbins = int(line[1])
                    break

        # Check axes names and relabel
        if np.array_equal(axes_names, ['"Initial energy [keV]"', '"#nu [deg]" "#lambda [deg]"', '"Polarization Angle [deg]"', '"Measured energy [keV]"', '"#phi [deg]"', '"#psi [deg]" "#chi [deg]"', '"#sigma [deg]" "#tau [deg]"', '"Distance [cm]"']):
            has_polarization = True
            labels = ("Ei", "NuLambda", "Pol", "Em", "Phi", "PsiChi", "SigmaTau", "Dist")
        elif np.array_equal(axes_names, ['"Initial energy [keV]"', '"#nu [deg]" "#lambda [deg]"', '"Measured energy [keV]"', '"#phi [deg]"', '"#psi [deg]" "#chi [deg]"', '"#sigma [deg]" "#tau [deg]"', '"Distance [cm]"']):
            has_polarization = False
            labels = ("Ei", "NuLambda", "Em", "Phi", "PsiChi", "SigmaTau", "Dist")
        else:
            raise InputError("Unknown response format")
        
        #check if the type of spectrum is known
        assert norm=="powerlaw" or norm=="Mono" or norm=="Linear" or norm=="Gaussian",f"unknown normalisation ! {norm}" 
         
        #check the number of simulated events is not 0
        assert nevents_sim != 0,"number of simulated events is 0 !" 
        
        
        logger.info("normalisation is {0}".format(norm))
        if sparse == None :
            logger.info("Sparse paramater not found in the file : We assume this is a non sparse matrice !")
            sparse = False
        else :
            logger.info("Sparse matrice ? {0}".format(sparse))
        edges = ()

        for axis_edges, axis_type in zip(axes_edges, axes_types):

            if axis_type == 'HEALPix':

                if axis_edges == -1:
                    # Single bin axis --i.e. all-sky
                    edges += ([0,1],)
                else:
                    edges += (np.arange(axis_edges+1),)

            elif axis_type == "FISBEL":
                raise RuntimeError("FISBEL binning not currently supported")
            else:
                edges += (axis_edges,)
        
        if sparse :
            axes = Axes(edges, labels=labels)
        
        else :
            axes = Axes(edges[:-2], labels=labels[:-2])            


        if sparse :
            # Need to get number of lines for progress bar.
            # First try fast method for unix-based systems:
            try:
                proc=subprocess.Popen('gunzip -c %s | wc -l' %filename, \
                        shell=True, stdout=subprocess.PIPE)
                nlines = int(proc.communicate()[0])


            # If fast method fails, use long method, which should work in all cases.
            except:
                logger.info("Initial attempt failed.")
                logger.info("Using long method...")
                nlines = sum(1 for _ in gzip.open(filename,"rt"))
                
            # Preallocate arrays
            coords = np.empty([axes.ndim, nlines], dtype=np.uint32)
            data = np.empty(nlines, dtype=np.uint32)

            # Calculate the memory usage in Gigabytes
            memory_size = ((nlines * data.itemsize)+(axes.ndim*nlines*coords.itemsize))/(1024*1024*1024)
            logger.info(f"Estimated RAM you need to read the file : {memory_size} GB")

    
                
        else :
            nlines = nbins        
            
            # Preallocate arrays    
            data = np.empty(nlines, dtype=np.uint32)

            # Calculate the memory usage in Gigabytes
            memory_size = (nlines * data.itemsize)/(1024*1024*1024)
            logger.info(f"Estimated RAM you need to read the file : {memory_size} GB")


        # Loop
        sbin = 0

        # read the rsp file and get the bin number and counts
        with gzip.open(filename, "rt") as file:
             


            #sparse case
            if sparse :
            
                progress_bar = tqdm(file, total=nlines, desc="Progress", unit="line")
                
                for line in progress_bar:

                
                    line = line.split()

                    if len(line) == 0:
                        continue

                    key = line[0]

                    if key == 'RD':

                        b = np.array(line[1:-1], dtype=np.uint32)
                        c = int(line[-1])

                        coords[:, sbin] = b
                        data[sbin] = c

                        sbin += 1
                    if sbin%10e6 == 0 : 
                        progress_bar.update(10e6)
            
                progress_bar.close()
                nbins_sparse = sbin

            #non sparse case
            else :
                

                 
                binLine = False 
                         
                for line in file:
                    line = line.split()
                    
                    if len(line) == 0:
                        continue
                    
                    if line[0] == "StartStream" :
                        binLine = True
                        continue
                        
                    if binLine :
                        #check we have same number of bin than values read
                        if len(line)!=nbins :
                            logger.info("nb of bin content read ({0}) != nb of bins {1}".format(len(line),nbins))
                            sys.exit()
                        
                        for i in tqdm(range(nbins), desc="Processing", unit="bin"):
                            data[i] = line[i]  
                
                        # we reshape the bincontent to the response matrice dimension
                        # note that for non sparse matrice SigmaTau and Dist are not used
                        data = np.reshape(data,tuple(axes.nbins),order="F")

                        break
        
        logger.info("response file read ! Now we create the histogram and weight in order to "+ 
                "get the effective area")
        # create histpy histogram

        
        if sparse :
            dr = Histogram(axes, contents=COO(coords=coords[:, :nbins_sparse], data= data[:nbins_sparse], shape = tuple(axes.nbins)))

        else :
        
            dr = Histogram(axes, contents=data)
        
        
        
	
	
        # Weight to get effective area

        ewidth = dr.axes['Ei'].widths
        ecenters = dr.axes['Ei'].centers
        
        #print(ewidth)
        #print(ecenters)

        #if we have one single bin, treat the gaussian norm like the mono one
        #also check that the gaussian spectrum is fully contained in that bin 
        if len(ewidth) == 1 and norm == "Gaussian":
            edges = dr.axes['Ei'].edges
            gauss_int = 0.5 * (1 + erf( (edges[0]-Gauss_mean)/(4*np.sqrt(2)) ) ) + 0.5 * (1 + erf( (edges[1]-Gauss_mean)/(4*np.sqrt(2)) ) )
            
            assert gauss_int == 1, "The gaussian spectrum is not fully contained in this single bin !"
            logger.info("Only one bin so we will use the Mono normalisation")
            norm ="Mono"

        if Spectrumfile is not None and norm=="file":
            logger.info("normalisation : spectrum file")
            # From spectrum file
            spec = pd.read_csv(Spectrumfile, sep=" ")
            spec = spec.iloc[:-1]
            hspec = Histogram([h_spec.axes[1]])
            hspec[:] = np.interp(hspec.axis.centers,
                             spec.iloc[:, 0].to_numpy(),
                             spec.iloc[:, 1].to_numpy(),
                             left=0,
                             right=0) * ewidth
            hspec /= np.sum(hspec)

            nperchannel_norm = hspec[:]

        elif norm=="powerlaw":
            logger.info("normalisation : powerlaw with index {0} with energy range [{1}-{2}]keV".format(alpha,emin,emax))
            # From powerlaw

            e_lo = dr.axes['Ei'].lower_bounds
            e_hi = dr.axes['Ei'].upper_bounds

            e_lo = np.minimum(emax, e_lo)
            e_hi = np.minimum(emax, e_hi)

            e_lo = np.maximum(emin, e_lo)
            e_hi = np.maximum(emin, e_hi)

            if alpha == 1:

                nperchannel_norm = np.log(e_hi/e_low) / np.log(emax/emin)
                
            else:

                a = 1 - alpha
                
                nperchannel_norm = (e_hi**a - e_lo**a) / (emax**a - emin**a)            

        elif norm =="Linear" :
            logger.info("normalisation : linear with energy range [{0}-{1}]".format(emin,emax))
            nperchannel_norm = ewidth / (emax-emin)
            
        elif norm=="Mono" :
            logger.info("normalisation : mono")

            nperchannel_norm = np.array([1.])
        
        elif norm == "Gaussian" :
            raise NotImplementedError("Gausssian norm for multiple bins not yet implemented")


        nperchannel = nperchannel_norm * nevents_sim
        # Full-sky?
        if not single_pixel:

            # Assumming all FISBEL pixels have the same area
            nperchannel /= dr.axes['NuLambda'].nbins

        # Area
        counts2area = area_sim / nperchannel
        dr_area = dr * dr.expand_dims(counts2area, 'Ei')

        

        #delete the array of data in order to release some memory
        del data 


        # end of weight now we create the .h5 structure

        npix = dr_area.axes['NuLambda'].nbins

        # remove the .h5 file if it already exist
        try:
            os.remove(filename.replace(".rsp.gz", "_nside{0}.area.h5".format(nside)))
        except:
            pass

        # create a .h5 file with the good structure
        filename = Path(str(filename).replace(".rsp.gz","_nside{0}.area.h5".format(nside)))
        f = h5.File(filename, mode='w')

        drm = f.create_group('DRM')

        # Header
        drm.attrs['UNIT'] = 'cm2'

        axis_description = {'Ei': "Initial simulated energy",
                            'NuLambda': "Location of the simulated source in the spacecraft coordinates",
                            'Pol': "Polarization angle",
                            'Em': "Measured energy",
                            'Phi': "Compton angle",
                            'PsiChi': "Location in the Compton Data Space",
                            'SigmaTau': "Electron recoil angle",
                            'Dist': "Distance from first interaction"
                            }

        #keep the same dimension order of the data
        axes_to_write = ['NuLambda', 'Ei']
        
        if has_polarization:
            axes_to_write += ['Pol']

        axes_to_write += ['Em', 'Phi', 'PsiChi']

        if sparse:
            drm.attrs['SPARSE'] = True
            
            # singletos. Save space in dense
            axes_to_write += ['SigmaTau', 'Dist']
        else:
            drm.attrs['SPARSE'] = False
            
        axes = drm.create_group('AXES', track_order=True)

        for axis in dr.axes[axes_to_write]:

            axis_dataset = axes.create_dataset(axis.label,
                                    data=axis.edges)


            if axis.label in ['NuLambda', 'PsiChi','SigmaTau']:

                # HEALPix
                axis_dataset.attrs['TYPE'] = 'healpix'

                axis_dataset.attrs['NSIDE'] = nside

                axis_dataset.attrs['SCHEME'] = 'ring'

            else:

                # 1D
                axis_dataset.attrs['TYPE'] = axis.axis_scale

                if axis.label in ['Ei', 'Em']:
                    axis_dataset.attrs['UNIT'] = 'keV'
                    axis_dataset.attrs['TYPE'] = 'log'
                elif axis.label in ['Phi', 'Pol']:
                    axis_dataset.attrs['UNIT'] = 'deg'
                    axis_dataset.attrs['TYPE'] = 'linear'
                elif axis.label in ['Dist']:
                    axis_dataset.attrs['UNIT'] = 'cm'
                    axis_dataset.attrs['TYPE'] = 'linear'
                else:
                   raise ValueError("Shouldn't happend")

            axis_dataset.attrs['DESCRIPTION'] = axis_description[axis.label]

        #sparse matrice
        if sparse :
        
            progress_bar = tqdm(total=npix, desc="Progress", unit="nbpixel")
            # Contents. Sparse arrays
            coords = drm.create_dataset('BIN_NUMBERS',
                                (npix,),
                                dtype=h5.vlen_dtype(int),
                                compression="gzip")

            data = drm.create_dataset('CONTENTS',
                              (npix,),
                              dtype=h5.vlen_dtype(float),
                              compression="gzip")
        

        


            for b in range(npix):
        
                #print(f"{b}/{npix}")
        
                pix_slice = dr_area[{'NuLambda':b}]
                
        
                coords[b] = pix_slice.coords.flatten()
                data[b] = pix_slice.data
                progress_bar.update(1)
            
            progress_bar.close()

        #non sparse
        else :
           
     
            data = drm.create_dataset('CONTENTS',
                              data=np.transpose(dr_area.contents, axes = [1,0,2,3,4]),
                              
                              compression="gzip")
        
            
                
                
        

        
        #close the .h5 file in write mode in order to reopen it in read mode after
        f.close()
        
        new = cls(filename)

        new._file = h5.File(filename, mode='r')
        new._drm = new._file['DRM']

        new._unit = u.Unit(new._drm.attrs['UNIT'])
        new._sparse = new._drm.attrs['SPARSE']
        

        # Axes
        axes = []

        for axis_label in new._drm["AXES"]:

            axis = new._drm['AXES'][axis_label]

            axis_type = axis.attrs['TYPE']



            if axis_type == 'healpix':

                axes += [HealpixAxis(edges=np.array(axis),
                                         nside=axis.attrs['NSIDE'],
                                         label=axis_label,
                                         scheme=axis.attrs['SCHEME'],
                                         coordsys=SpacecraftFrame())]

            else:
                axes += [Axis(np.array(axis) * u.Unit(axis.attrs['UNIT']),
                                  scale=axis_type,
                                  label=axis_label)]

                

        new._axes = Axes(axes)

        # Init HealpixMap (local coordinates, main axis)
        HealpixBase.__init__(new,
                                 base=new.axes['NuLambda'],
                                 coordsys=SpacecraftFrame())

        return new

    @property
    def is_sparse(self):
        return self._sparse

    @property
    def ndim(self):
        """
        Dimensionality of detector response matrix.

        Returns
        -------
        int
        """

        return self.axes.ndim

    @property
    def axes(self):
        """
        List of axes.

        Returns
        -------
        :py:class:`histpy.Axes`
        """
        return self._axes

    @property
    def unit(self):
        """
        Physical unit of the contents of the detector reponse.

        Returns
        -------
        :py:class:`astropy.units.Unit`
        """

        return self._unit

    def __getitem__(self, pix):

        if not isinstance(pix, (int, np.integer)) or pix < 0 or not pix < self.npix:
            raise IndexError("Pixel number out of range, or not an integer")

        #check if we have sparse matrice or not

        if self._sparse:
            coords = np.reshape(
                self._file['DRM']['BIN_NUMBERS'][pix], (self.ndim-1, -1))
            data = np.array(self._file['DRM']['CONTENTS'][pix])

            return DetectorResponse(self.axes[1:],
                                contents=COO(coords=coords,
                                             data=data,
                                             shape=tuple(self.axes.nbins[1:])),
                                unit=self.unit)
                                
        else :
            data = self._file['DRM']['CONTENTS'][pix]
            return DetectorResponse(self.axes[1:],
                                contents=data, unit=self.unit)                 

    def close(self):
        """
        Close the HDF5 file containing the response
        """

        self._file.close()

    def __enter__(self):
        """
        Start a context manager
        """

        return self

    def __exit__(self, type, value, traceback):
        """
        Exit a context manager
        """

        self.close()

    @property
    def filename(self):
        """
        Path to on-disk file containing DetectorResponse

        Returns
        -------
        :py:class:`~pathlib.Path`
        """

        return Path(self._file.filename)

    def get_interp_response(self, coord):
        """
        Get the bilinearly interpolated response at a given coordinate location.

        Parameters
        ----------
        coord : :py:class:`astropy.coordinates.SkyCoord`
            Coordinate in the :py:class:`.SpacecraftFrame`

        Returns
        -------
        :py:class:`DetectorResponse`
        """

        pixels, weights = self.get_interp_weights(coord)

        

        dr = DetectorResponse(self.axes[1:],
                              sparse=self._sparse,
                              unit=self.unit)
        
        
        for p, w in zip(pixels, weights):

            dr += self[p]*w

        return dr

    def get_point_source_response(self,
                                  exposure_map = None,
                                  coord = None,
                                  scatt_map = None,
                                  Earth_occ = True):
        """
        Convolve the all-sky detector response with exposure for a source at a given
        sky location.

        Provide either a exposure map (aka dweel time map) or a combination of a 
        sky coordinate and a spacecraft attitude map.

        Parameters
        ----------
        exposure_map : :py:class:`mhealpy.HealpixMap`
            Effective time spent by the source at each pixel location in spacecraft coordinates
        coord : :py:class:`astropy.coordinates.SkyCoord`
            Source coordinate
        scatt_map : :py:class:`SpacecraftAttitudeMap`
            Spacecraft attitude map
        Earth_occ : bool, optional
            Option to include Earth occultation in the respeonce. 
            Default is True, in which case you can only pass one 
            coord, which must be the same as was used for the scatt map. 
        
        Returns
        -------
        :py:class:`PointSourceResponse`
        """

        # TODO: deprecate exposure_map in favor of coords + scatt map for both local
        # and interntial coords
        
        if Earth_occ == True:
            if coord != None:
                if coord.size > 1:
                    raise ValueError("For Earth occultation you must use the same coordinate as was used for the scatt map!")

        if exposure_map is not None:
            if not self.conformable(exposure_map):
                raise ValueError(
                    "Exposure map has a different grid than the detector response")

            psr = PointSourceResponse(self.axes[1:],
                                      sparse=self._sparse,
                                      unit=u.cm*u.cm*u.s)

            for p in range(self.npix):

                if exposure_map[p] != 0:
                    psr += self[p]*exposure_map[p]

            return psr

        else:

            # Rotate to inertial coordinates

            if coord is None or scatt_map is None:
                raise ValueError("Provide either exposure map or coord + scatt_map")
            
            if isinstance(coord.frame, SpacecraftFrame):
                raise ValueError("Local coordinate + scatt_map not currently supported")

            if self.is_sparse:
                raise ValueError("Coord +  scatt_map currently only supported for dense responses")

            axis = "PsiChi"

            coords_axis = Axis(np.arange(coord.size+1), label = 'coords')

            psr = Histogram([coords_axis] + list(deepcopy(self.axes[1:])), 
                            unit = self.unit * scatt_map.unit)
            
            psr.axes[axis].coordsys = coord.frame

            for i,(pixels, exposure) in \
                enumerate(zip(scatt_map.contents.coords.transpose(),
                              scatt_map.contents.data)):

                #gc.collect() # HDF5 cache issues
                
                att = Attitude.from_axes(x = scatt_map.axes['x'].pix2skycoord(pixels[0]),
                                         y = scatt_map.axes['y'].pix2skycoord(pixels[1]))

                coord.attitude = att

                #TODO: Change this to interpolation
                loc_nulambda_pixels = np.array(self.axes['NuLambda'].find_bin(coord),
                                               ndmin = 1)
                
                dr_pix = Histogram.concatenate(coords_axis, [self[i] for i in loc_nulambda_pixels])

                dr_pix.axes['PsiChi'].coordsys = SpacecraftFrame(attitude = att)

                self._sum_rot_hist(dr_pix, psr, exposure)

            # Convert to PSR
            psr = tuple([PointSourceResponse(psr.axes[1:],
                                             contents = data,
                                             sparse = psr.is_sparse,
                                             unit = psr.unit)
                         for data in psr[:]])
            
            if coord.size == 1:
                return psr[0]
            else:
                return psr

    def _setup_extended_source_response_params(self, coordsys, nside_image, nside_scatt_map):
        """
        Validate coordinate system and setup NSIDE parameters for extended source response generation.

        Parameters
        ----------
        coordsys : str
            Coordinate system to be used (currently only 'galactic' is supported)
        nside_image : int or None
            NSIDE parameter for the image reconstruction.
            If None, uses the full detector response's NSIDE.
        nside_scatt_map : int or None
            NSIDE parameter for scatt map generation.
            If None, uses the full detector response's NSIDE.

        Returns
        -------
        tuple
            (coordsys, nside_image, nside_scatt_map) : validated parameters
        """
        if coordsys != 'galactic':
            raise ValueError(f'The coordsys {coordsys} not currently supported')

        if nside_image is None:
            nside_image = self.nside

        if nside_scatt_map is None:
            nside_scatt_map = self.nside
            
        return coordsys, nside_image, nside_scatt_map

    def get_point_source_response_per_image_pixel(self, ipix_image, orientation, coordsys = 'galactic', nside_image = None, nside_scatt_map = None, Earth_occ = True):
        """
        Generate point source response for a specific HEALPix pixel by convolving 
        the all-sky detector response with exposure.

        Parameters
        ----------
        ipix_image : int
            HEALPix pixel index
        orientation : cosipy.spacecraftfile.SpacecraftFile
            Spacecraft attitude information
        coordsys : str, default 'galactic'
            Coordinate system (currently only 'galactic' is supported)
        nside_image : int, optional
            NSIDE parameter for image reconstruction.
            If None, uses the detector response's NSIDE.
        nside_scatt_map : int, optional
            NSIDE parameter for scatt map generation.
            If None, uses the detector response's NSIDE.
        Earth_occ : bool, default True
            Whether to include Earth occultation in the response

        Returns
        -------
        :py:class:`PointSourceResponse`
            Point source response for the specified pixel
        """
        coordsys, nside_image, nside_scatt_map = self._setup_extended_source_response_params(coordsys, nside_image, nside_scatt_map)
        
        image_axes = HealpixAxis(nside = nside_image, coordsys = coordsys, scheme='ring', label = 'NuLambda') # The label should be 'lb' in the future

        coord = image_axes.pix2skycoord(ipix_image)

        scatt_map = orientation.get_scatt_map(target_coord = coord,
                                              nside = nside_scatt_map,
                                              scheme='ring',
                                              coordsys=coordsys,
                                              earth_occ=Earth_occ)

        psr = self.get_point_source_response(coord = coord, scatt_map = scatt_map, Earth_occ = Earth_occ)

        return psr

    def get_extended_source_response(self, orientation, coordsys = 'galactic', nside_image = None, nside_scatt_map = None, Earth_occ = True):
        """
        Generate extended source response by convolving the all-sky detector 
        response with exposure over the entire sky.

        Parameters
        ----------
        orientation : cosipy.spacecraftfile.SpacecraftFile
            Spacecraft attitude information
        coordsys : str, default 'galactic'
            Coordinate system (currently only 'galactic' is supported)
        nside_image : int, optional
            NSIDE parameter for image reconstruction.
            If None, uses the detector response's NSIDE.
        nside_scatt_map : int, optional
            NSIDE parameter for scatt map generation.
            If None, uses the detector response's NSIDE.
        Earth_occ : bool, default True
            Whether to include Earth occultation in the response

        Returns
        -------
        :py:class:`ExtendedSourceResponse`
            Extended source response covering the entire sky
        """
        coordsys, nside_image, nside_scatt_map = self._setup_extended_source_response_params(coordsys, nside_image, nside_scatt_map)

        axes = [HealpixAxis(nside = nside_image, coordsys = coordsys, scheme='ring', label = 'NuLambda')] # The label should be 'lb' in the future
        axes += list(self.axes[1:])
        axes[-1].coordsys = coordsys

        extended_source_response = ExtendedSourceResponse(axes, unit = u.Unit("cm2 s"))

        for ipix in tqdm(range(hp.nside2npix(nside_image))):
    
            psr = self.get_point_source_response_per_image_pixel(ipix, orientation, coordsys = coordsys, 
                                                                 nside_image = nside_image, nside_scatt_map = nside_scatt_map, Earth_occ = Earth_occ)

            extended_source_response[ipix] = psr.contents

        return extended_source_response

    def merge_psr_to_extended_source_response(self, basename, coordsys = 'galactic', nside_image = None):
        """
        Create extended source response by merging multiple point source responses.
    
        Reads point source response files matching the pattern `basename` + index + file_extension. 
        For example, with basename='histograms/hist_', filenames are expected to be like 'histograms/hist_00001.hdf5'.

        Parameters
        ----------
        basename : str
            Base filename pattern for point source response files
        coordsys : str, default 'galactic'
            Coordinate system (currently only 'galactic' is supported)
        nside_image : int, optional
            NSIDE parameter for image reconstruction.
            If None, uses the detector response's NSIDE.

        Returns
        -------
        :py:class:`ExtendedSourceResponse`
            Combined extended source response
        """
        coordsys, nside_image, _ = self._setup_extended_source_response_params(coordsys, nside_image, None)

        psr_files = glob.glob(basename + "*")

        if not psr_files:
            raise FileNotFoundError(f"No files found matching pattern {basename}*")

        axes = [HealpixAxis(nside = nside_image, coordsys = coordsys, scheme='ring', label = 'NuLambda')] # The label should be 'lb' in the future
        axes += list(self.axes[1:])
        axes[-1].coordsys = coordsys

        extended_source_response = ExtendedSourceResponse(axes, unit = u.Unit("cm2 s"))
        
        filled_pixels = []

        for filename in psr_files:

            ipix = int(filename[len(basename):].split(".")[0])

            psr = Histogram.open(filename)
        
            extended_source_response[ipix] = psr.contents

            filled_pixels.append(ipix)

        expected_pixels = set(range(extended_source_response.axes[0].npix))
        if set(filled_pixels) != expected_pixels:
            raise ValueError(f"Missing pixels in the response files. Expected {extended_source_response.axes[0].npix} pixels, got {len(filled_pixels)} pixels")

        return extended_source_response
            
    @staticmethod
    def _sum_rot_hist(h, h_new, exposure, axis = "PsiChi"):
        """
        Rotate a histogram with HealpixAxis h into the grid of h_new, and sum
        it up with the weight of exposure.

        Meant to rotate the PsiChi of a CDS from local to galactic
        """
        
        axis_id = h.axes.label_to_index(axis)

        old_axes = h.axes
        new_axes = h_new.axes

        old_axis = h.axes[axis_id]
        new_axis = h_new.axes[axis_id]

        # Convolve
        # TODO: Change this to interpolation (pixels + weights)
        old_pixels = old_axis.find_bin(new_axis.pix2skycoord(np.arange(new_axis.nbins)))
        # NOTE: there are some pixels that are duplicated, since the center 2 pixels
        # of the original grid can land within the boundaries of a single pixel
        # of the target grid. The following commented code fixes this, but it's slow, and
        # the effect is very small, so maybe not worth it
        # nulambda_npix = h.axes['NuLamnda'].nbins    
        # new_norm = np.zeros(shape = nulambda_npix)
        # for p in old_pixels:
        #     h_slice = h[{axis:p}]
        #     norm_rot += np.sum(h_slice, axis = tuple(np.arange(1, h_slice.ndim)))
        # old_norm = np.sum(h, axis = tuple(np.arange(1, h.ndim)))
        # norm_corr = h.expand_dims(norm / norm_rot, "NuLambda")

        for old_pix,new_pix in zip(old_pixels,range(new_axis.npix)):

            #h_new[{axis:new_pix}] += exposure * h[{axis: old_pix}] # * norm_corr
            # The following code does the same than the code above, but is faster
            # However, it uses some internal functionality in histpy, which is bad practice
            # TODO: change this in a future version. We might need to modify histpy so that
            # this is not needed
            
            old_indices = tuple([slice(None)]*axis_id + [old_pix+1])
            new_indices = tuple([slice(None)]*axis_id + [new_pix+1])

            h_new._contents[new_indices] += exposure * h._contents[old_indices] # * norm_corr
                        

    def __str__(self):
        return f"{self.__class__.__name__}(filename = '{self.filename.resolve()}')"

    def __repr__(self):

        output = (f"FILENAME: '{self.filename.resolve()}'\n"
                  f"AXES:\n")

        for naxis, axis in enumerate(self.axes):

            if naxis == 0:
                description = "Location of the simulated source in the spacecraft coordinates"
            else:
                description = self._drm['AXES'][axis.label].attrs['DESCRIPTION']

            output += (f"  {axis.label}:\n"
                       f"    DESCRIPTION: '{description}'\n")

            if isinstance(axis, HealpixAxis):
                output += (f"    TYPE: 'healpix'\n"
                           f"    NPIX: {axis.npix}\n"
                           f"    NSIDE: {axis.nside}\n"
                           f"    SCHEME: '{axis.scheme}'\n")
            else:
                output += (f"    TYPE: '{axis.axis_scale}'\n"
                           f"    UNIT: '{axis.unit}'\n"
                           f"    NBINS: {axis.nbins}\n"
                           f"    EDGES: [{', '.join([str(e) for e in axis.edges])}]\n")

        return output

    def _repr_pretty_(self, p, cycle):

        if cycle:
            p.text(str(self))
        else:
            p.text(repr(self))


def cosi_response(argv=None):
    """
    Print the content of a detector response to stdout.
    """

    # Parse arguments from commandline
    apar = argparse.ArgumentParser(
        usage=textwrap.dedent(
            """
            %(prog)s [--help] <command> [<args>] <filename> [<options>]
            """),
        description=textwrap.dedent(
            """
            Quick view of the information contained in a response file

            %(prog)s --help
            %(prog)s dump header [FILENAME]
            %(prog)s dump aeff [FILENAME] --lon [LON] --lat [LAT]
            %(prog)s dump expectation [FILENAME] --config [CONFIG]
            %(prog)s plot aeff [FILENAME] --lon [LON] --lat [LAT]
            %(prog)s plot dispersion [FILENAME] --lon [LON] --lat [LAT]
            %(prog)s plot expectation [FILENAME] --lon [LON] --lat [LAT]

            Arguments:
            - header: Response header and axes information
            - aeff: Effective area
            - dispersion: Energy dispection matrix
            - expectation: Expected number of counts
            """),
        formatter_class=argparse.RawTextHelpFormatter)

    apar.add_argument('command',
                      help=argparse.SUPPRESS)
    apar.add_argument('args', nargs='*',
                      help=argparse.SUPPRESS)
    apar.add_argument('filename',
                      help="Path to instrument response")
    apar.add_argument('--lon',
                      help="Longitude in sopacecraft coordinates. e.g. '11deg'")
    apar.add_argument('--lat',
                      help="Latitude in sopacecraft coordinates. e.g. '10deg'")
    apar.add_argument('--output', '-o',
                      help="Save output to file. Default: stdout")
    apar.add_argument('--config', '-c',
                      help="Path to config file describing exposure and source charateristics.")
    apar.add_argument('--config-override', dest='override',
                      help="Override option in config file")

    args = apar.parse_args(argv)

    # Config
    if args.config is None:
        config = Configurator()
    else:
        config = Configurator.open(args.config)

        if args.override is not None:
            config.override(args.override)

    # Get info
    with FullDetectorResponse.open(args.filename) as response:

        # Commands and functions
        def get_drm():

            lat = Quantity(args.lat)
            lon = Quantity(args.lon)

            loc = SkyCoord(lon=lon, lat=lat, frame=SpacecraftFrame())

            return response.get_interp_response(loc)

        def get_expectation():

            # Exposure map
            exposure_map = HealpixMap(base=response,
                                      unit=u.s,
                                      coordsys=SpacecraftFrame())

            ti = Time(config['exposure:time_i'])
            tf = Time(config['exposure:time_f'])
            dt = (tf-ti).to(u.s)

            exposure_map[:4] = dt/4

            logger.warning(f"Spacecraft file not yet implemented, faking source on "
                           f"axis from {ti} to {tf} ({dt:.2f})")

            # Point source response
            psr = response.get_point_source_response(exposure_map)

            # Spectrum
            model = ModelParser(model_dict=config['sources']).get_model()

            spectrum = model.point_sources['source'].components['main'].shape
            logger.info(f"Using spectrum:\n {spectrum}")

            # Expectation
            expectation = psr.get_expectation(spectrum).project('Em')

            return expectation

        def command_dump():

            if len(args.args) != 1:
                apar.error("Command 'dump' takes a single argument")

            option = args.args[0]

            if option == 'header':

                result = repr(response)

            elif option == 'aeff':

                drm = get_drm()

                aeff = drm.get_spectral_response().get_effective_area()

                result = "#Energy[keV]     Aeff[cm2]\n"

                for e, a in zip(aeff.axis.centers, aeff):
                    # IMC: fix this latter when histpy has units
                    result += f"{e.to_value(u.keV):>12.2e}  {a.to_value(u.cm*u.cm):>12.2e}\n"

            elif option == 'expectation':

                expectation = get_expectation()

                result = "#Energy_min[keV]   Energy_max[keV]  Expected_counts\n"

                for emin, emax, ex in zip(expectation.axis.lower_bounds,
                                          expectation.axis.upper_bounds,
                                          expectation):
                    # IMC: fix this latter when histpy has units
                    result += (f"{emin.to_value(u.keV):>16.2e}  "
                               f"{emax.to_value(u.keV):>16.2e}  "
                               f"{ex:>15.2e}\n")

            else:

                apar.error(f"Argument '{option}' not valid for 'dump' command")

            if args.output is None:
                logger.info(result)
            else:
                logger.info(f"Saving result to {Path(args.output).resolve()}")
                f = open(args.output, 'a')
                f.write(result)
                f.close()

        def command_plot():

            if len(args.args) != 1:
                apar.error("Command 'plot' takes a single argument")

            option = args.args[0]

            if option == 'aeff':

                drm = get_drm()

                drm.get_spectral_response().get_effective_area().plot(errorbars=False)

            elif option == 'dispersion':

                drm = get_drm()

                drm.get_spectral_response().get_dispersion_matrix().plot()

            elif option == 'expectation':

                expectation = get_expectation().plot(errorbars=False)

            else:

                apar.error(f"Argument '{option}' not valid for 'plot' command")

            if args.output is None:
                plt.show()
            else:
                logger.info(f"Saving plot to {Path(args.output).resolve()}")
                plt.savefig(args.output)

        # Run
        if args.command == 'plot':
            command_plot()
        elif args.command == 'dump':
            command_dump()
        else:
            apar.error(f"Command '{args.command}' unknown")



import logging
logger = logging.getLogger(__name__)

import numpy as np

from copy import deepcopy

from histpy import Histogram, Axes, Axis

import astropy.units as u

class DetectorResponse(Histogram):
    """
    Handles the multi-dimensional matrix that describes the
    response of the instrument for a particular :py:class:`.SpacecraftFrame` coordinate
    location.

    Parameters
    ----------
    axes : :py:class:`histpy.Axes`
        Binning information for each variable. The following labels are expected:\n
        - ``Ei``: Real energy
        - ``Em``: Measured energy
        - ``Phi``: Compton angle. Optional.
        - ``PsiChi``:  Location in the Compton Data Space (HEALPix pixel). Optional.
        - ``SigmaTau``: Electron recoil angle (HEALPix pixel). Optional.
        - ``Dist``: Distance from first interaction. Optional.
    contents : array, :py:class:`astropy.units.Quantity` or :py:class:`sparse.SparseArray`
        Array containing the differential effective area.
    unit : :py:class:`astropy.units.Unit`, optional
        Physical area units, if not specified as part of ``contents``
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._spec = None
        self._aeff = None
    
    def get_spectral_response(self, copy = True):
        """
        Reduced detector response, projected along the real and measured energy axes only.
        The Compton Data Space axes are not included.

        Parameters
        ----------
        copy : bool
            If true, a copy of the cached spectral response will be returned.
        
        Returns
        -------
        :py:class:`DetectorResponse`
        """

        # Cache the spectral response
        if self._spec is None:
            spec = self.project(['Ei','Em'])
            self._spec = DetectorResponse(spec.axes,
                                          contents = spec.contents,
                                          unit = spec.unit)

        if copy:
            return deepcopy(self._spec)
        else:
            return self._spec
        
    def get_effective_area(self, energy = None, copy = True):
        """
        Compute the effective area at a given energy. If no energy is specified, the
        output is a histogram for the effective area at each energy bin.

        Parameters
        ----------
        energy : optional, :py:class:`astropy.units.Quantity`
            Energy/energies at which to interpolate the linearly effective area
        copy : bool
            If true, a copy of the cached effective will be returned.
        
        Returns
        -------
        :py:class:`astropy.units.Quantity` or :py:class:`histpy.Histogram`
        """
        
        if self._aeff is None:
            self._aeff = self.get_spectral_response(copy = False).project('Ei').to_dense()

        if energy is None:
            if copy:
                return deepcopy(self._aeff)
            else:
                return self._aeff
        else:
            return self._aeff.interp(energy)

    def get_dispersion_matrix(self):
        """
        Compute the energy dispersion matrix, also known as migration matrix. This holds the
        probability of an event with real energy ``Ei`` to be reconstructed with an measured
        energy ``Em``.

        Returns
        -------
        :py:class:`histpy.Histogram`
        """
        
        # Get spectral response and effective area normalization
        spec = self.get_spectral_response(copy = False)
        norm = self.get_effective_area().full_contents

        # Hack the under/overflow bins to supress 0/0 wearning
        norm[0] = 1*norm.unit if norm[0] == 0 else norm[0]
        norm[-1] = 1*norm.unit if norm[-1] == 0 else norm[-1]

        # Avoid another 0/0 is the effective area is null for some bins
        if np.any(norm == 0):
            norm[norm == 0] = 1*norm.unit

            logger.warn("Null effective area, cannot properly compute dispersion matrix.")
        
        # "Broadcast" such that it has the compatible dimensions with the 2D matrix
        norm = spec.expand_dims(norm, 'Ei')
        
        # Normalize column-by-column
        return (spec / norm)

    @property
    def photon_energy_axis(self):
        """
        Real energy bins (``Ei``).

        Returns
        -------
        :py:class:`histpy.Axes`
        """
        
        return self.axes['Ei']

    
    @property
    def measured_energy_axis(self):
        """
        Measured energy bins (``Em``).

        Returns
        -------
        :py:class:`histpy.Axes`        
        """
        
        return self.axes['Em']
        

        
    



import numpy as np
from astropy.coordinates import SkyCoord, Angle
import astropy.units as u

from .conventions import PolarizationConvention

class PolarizationAngle:

    def __init__(self, angle, skycoord ,
                 convention = 'iau',
                 *args, **kwargs):
        """
        Defines a polarization angle in the context of a source direction and
        polarization angle convention.

        Parameters:
        angle : :py:class:`astropy.coordinates.Angle
            Polarization angle
        skycoord : :py:class:`astropy.coordinates.SkyCoord`
            Source direction
        convention : PolarizationConvention
            Convention the defined the polarization basis and direction in 
            the polarization plane (for which the source direction is normal)
        *args, **kwargs
            Passed to convention class.
        """

        # Ensure pa is an Angle object
        self._angle = Angle(angle)

        self._convention = PolarizationConvention.get_convention(convention,
                                                                 *args, **kwargs)

        self._skycoord = skycoord

    def __repr__(self):
        return f"<PolarizationAngle({self._angle.degree} deg at {self._skycoord} using convention {self._convention})>"
        
    @property
    def angle(self):
        return self._angle

    @property
    def convention(self):
        return self._convention

    @property
    def skycoord(self):
        return self._skycoord

    @property
    def vector(self):
        """
        Direction of the electric field vector
        """

        # Get the projection vectors for the source direction in the current convention
        px, py = self.convention.get_basis(self.skycoord)

        px = px.cartesian.xyz
        py = py.cartesian.xyz

        # Calculate the cosine and sine of the polarization angle
        cos_pa = np.cos(self.angle.radian)
        sin_pa = np.sin(self.angle.radian)

        # Calculate the polarization vector
        pol_vec = px * cos_pa + py * sin_pa
                
        return SkyCoord(*pol_vec,
                        representation_type = 'cartesian',
                        frame = self.convention.frame)
    
    def transform_to(self, convention, *args, **kwargs):

        # Standarize convention 2
        convention2 = PolarizationConvention.get_convention(convention, *args, **kwargs)
        
        # Calculate the polarization vector in the current convention
        pol_vec = self.vector.transform_to(convention2.frame).cartesian.xyz

        # Get the projection vectors for the source direction in the new convention
        (px2, py2) = convention2.get_basis(self.skycoord)

        px2 = px2.cartesian.xyz
        py2 = py2.cartesian.xyz
        
        # Compute the dot products for the transformation
        a = np.sum(pol_vec * px2, axis=0)
        b = np.sum(pol_vec * py2, axis=0)

        # Calculate the new polarization angle in the new convention
        pa_2 = Angle(np.arctan2(b, a), unit=u.rad)

        # Normalize the angle to be between 0 and pi
        if pa_2 < 0:
            pa_2 += Angle(np.pi, unit=u.rad)

        return PolarizationAngle(pa_2,
                                 self.skycoord,
                                 convention = convention2)

        



from histpy import Histogram, Axes, Axis
import numpy as np
import astropy.units as u
import gc

from .functions import get_integrated_extended_model

class ExtendedSourceResponse(Histogram):
    """
    A class to represent and manipulate extended source response data.

    This class provides methods to load data from HDF5 files, access contents,
    units, and axes information, and calculate expectations based on sky models.

    Methods
    -------
    get_expectation(allsky_image_model)
        Calculate expectation based on an all-sky image model.
    get_expectation_from_astromodel(source)
        Calculate expectation from an astronomical model source.

    Notes
    -----
    Currently, the axes of the response must be ['NuLambda', 'Ei', 'Em', 'Phi', 'PsiChi'].
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize an ExtendedSourceResponse object.
        """
        # Not to track the under/overflow bins
        kwargs['track_overflow'] = False

        super().__init__(*args, **kwargs)
        
        if not np.all(self.axes.labels == ['NuLambda', 'Ei', 'Em', 'Phi', 'PsiChi']):
            # 'NuLambda' should be 'lb' if it is in the gal. coordinates?
            raise ValueError(f"The input axes {self.axes.labels} is not supported by ExtendedSourceResponse class.")

    @classmethod
    def open(cls, filename, name='hist'):
        """
        Load data from an HDF5 file.

        Parameters
        ----------
        filename : str
            The path to the HDF5 file.
        name : str, optional
            The name of the histogram group in the HDF5 file (default is 'hist').

        Returns
        -------
        ExtendedSourceResponse
            A new instance of ExtendedSourceResponse with loaded data.

        Raises
        ------
        ValueError
            If the shape of the contents does not match the axes.
        """
        hist = super().open(filename, name)

        axes = hist.axes
        contents = hist[:]
        sumw2 = hist.sumw2 
        unit = hist.unit
        track_overflow = False
        
        new = cls(axes, contents = contents,
                        sumw2 = sumw2,
                        unit = unit,
                        track_overflow = track_overflow)

        if new.is_sparse:
            new = new.to_dense()
        
        del hist
        gc.collect()

        return new

    def get_expectation(self, allsky_image_model):
        """
        Calculate expectation based on an all-sky image model.

        Parameters
        ----------
        allsky_image_model : Histogram 
            The all-sky image model to use for calculation.

        Returns
        -------
        Histogram
            A histogram representing the calculated expectation.
        """
        if self.axes[0].label == allsky_image_model.axes[0].label \
            and self.axes[1].label == allsky_image_model.axes[1].label \
            and np.all(self.axes[0].edges == allsky_image_model.axes[0].edges) \
            and np.all(self.axes[1].edges == allsky_image_model.axes[1].edges) \
            and allsky_image_model.unit == u.Unit('1/(s*cm*cm*sr)'):
            
            contents = np.tensordot(allsky_image_model.contents, self.contents, axes=([0,1], [0,1]))
            contents *= self.axes[0].pixarea()

            return Histogram(edges=self.axes[2:], contents=contents)
        
        else:
            raise ValueError(f"The input allskymodel mismatches with the extended source response.")

    def get_expectation_from_astromodel(self, source):
        """
        Calculate expectation from an astromodels extended source model.

        This method creates an AllSkyImageModel based on the current axes configuration,
        sets its values from the provided astromodels extended source model, and then
        calculates the expectation using the get_expectation method.

        Parameters
        ----------
        source : astromodels.ExtendedSource
            An astromodels extended source model object. This model represents
            the spatial and spectral distribution of an extended astronomical source.

        Returns
        -------
        Histogram
            A histogram representing the calculated expectation based on the
            provided extended source model.
        """

        allsky_image_model = get_integrated_extended_model(source, image_axis = self.axes[0], energy_axis = self.axes[1])

        return self.get_expectation(allsky_image_model)



from .polarization_asad import PolarizationASAD, calculate_uncertainties
from .conventions import PolarizationConvention, OrthographicConvention, StereographicConvention, IAUPolarizationConvention
from .polarization_angle import PolarizationAngle



import numpy as np
from tqdm.autonotebook import tqdm
import astropy.units as u

import logging
logger = logging.getLogger(__name__)

import warnings

from histpy import Histogram, Axes

from cosipy.response import FullDetectorResponse

from .image_deconvolution_data_interface_base import ImageDeconvolutionDataInterfaceBase

class DataIF_COSI_DC2(ImageDeconvolutionDataInterfaceBase):
    """
    A subclass of ImageDeconvolutionDataInterfaceBase for the COSI data challenge 2.
    """

    def __init__(self, name = None):

        ImageDeconvolutionDataInterfaceBase.__init__(self, name)

        self._image_response = None # histpy.Histogram (dense)

        # None if using Galactic CDS, mandotary if using local CDS
        self._coordsys_conv_matrix = None 

        # optional
        self.is_miniDC2_format = False #should be removed in the future

    @classmethod
    def load(cls, name, event_binned_data, dict_bkg_binned_data, rsp, coordsys_conv_matrix = None, is_miniDC2_format = False):
        """
        Load data

        Parameters
        ----------
        name : str
            The name of data
        event_binned_data : :py:class:`histpy.Histogram`
            Event histogram
        dict_bkg_binned_data : dict
            Background models as {background_model_name: :py:class:`histpy.Histogram`}
        rsp : :py:class:`histpy.Histogram` or :py:class:`cosipy.response.FullDetectorResponse`
            Response
        coordsys_conv_matrix : :py:class:`cosipy.image_deconvolution.CoordsysConversionMatrix`, default False
            Coordsys conversion matrix 
        is_miniDC2_format : bool, default False
            Whether the file format is for mini-DC2. It will be removed in the future.

        Returns
        -------
        :py:class:`cosipy.image_deconvolution.DataIF_COSI_DC2`
            An instance of DataIF_COSI_DC2 containing the input data set
        """

        new = cls(name)

        new._event = event_binned_data.to_dense()

        new._bkg_models = dict_bkg_binned_data

        for key in new._bkg_models:
            if new._bkg_models[key].is_sparse:
                new._bkg_models[key] = new._bkg_models[key].to_dense()

            new._summed_bkg_models[key] = np.sum(new._bkg_models[key])

        new._coordsys_conv_matrix = coordsys_conv_matrix

        new.is_miniDC2_format = is_miniDC2_format

        if isinstance(rsp, FullDetectorResponse):
            logger.info('Loading the response matrix onto your computer memory...')
            new._load_full_detector_response_on_memory(rsp, is_miniDC2_format)
            logger.info('Finished')
        elif isinstance(rsp, Histogram):
            new._image_response = rsp
        
        # We modify the axes in event, bkg_models, response. This is only for DC2.
        new._modify_axes()
        
        new._data_axes = new._event.axes
        
        if new._coordsys_conv_matrix is None:
            axes = [new._image_response.axes['NuLambda'], new._image_response.axes['Ei']]
            axes[0].label = 'lb' 
            # The gamma-ray direction of pre-computed response in DC2 is in the galactic coordinate, not in the local coordinate.
            # Actually, it is labeled as 'NuLambda'. So I replace it with 'lb'.
            new._model_axes = Axes(axes)
        else:
            new._model_axes = Axes([new._coordsys_conv_matrix.axes['lb'], new._image_response.axes['Ei']])

        new._calc_exposure_map()

        return new

    def _modify_axes(self):
        """
        Modify the axes of data. This method will be removed in the future.
        """

        warnings.warn("Note that _modify_axes() in DataIF_COSI_DC2 was implemented for a temporary use. It will be removed in the future.", DeprecationWarning)

        if self._coordsys_conv_matrix is None:
            axis_name = ['Em', 'Phi', 'PsiChi']

        elif self._coordsys_conv_matrix.binning_method == 'Time':
            axis_name = ['Time', 'Em', 'Phi', 'PsiChi']

        elif self._coordsys_conv_matrix.binning_method == 'ScAtt':
            axis_name = ['ScAtt', 'Em', 'Phi', 'PsiChi']

        for name in axis_name:

            logger.info(f"... checking the axis {name} of the event and background files...")
            
            event_edges, event_unit = self._event.axes[name].edges, self._event.axes[name].unit

            for key in self._bkg_models:

                bkg_edges, bkg_unit = self._bkg_models[key].axes[name].edges, self._bkg_models[key].axes[name].unit

                if np.all(event_edges == bkg_edges):
                    logger.info(f"    --> pass (edges)") 
                else:
                    logger.error(f"Warning: the edges of the axis {name} are not consistent between the event and the background model {key}!")
                    logger.error(f"         event      : {event_edges}")
                    logger.error(f"         background : {bkg_edges}")
                    raise ValueError

        # check the axes of the event/response files. 
        # Note that currently (2023-08-29) no unit is stored in the binned data. So only the edges are compared. This should be modified in the future.

        axis_name = ['Em', 'Phi', 'PsiChi']
        
        for name in axis_name:

            logger.info(f"...checking the axis {name} of the event and response files...")

            event_edges, event_unit = self._event.axes[name].edges, self._event.axes[name].unit
            response_edges, response_unit = self._image_response.axes[name].edges, self._image_response.axes[name].unit
            
#            if type(response_edges) == u.quantity.Quantity and self.is_miniDC2_format == True:
            if event_unit is None and response_unit is not None and self.is_miniDC2_format == True: # this is only for the old data in the miniDC2 challenge. I will remove them in the near future (or in the final dataIF).
                response_edges = response_edges.value

            if np.all(event_edges == response_edges):
                logger.info(f"    --> pass (edges)") 
            else:
                logger.error(f"Warning: the edges of the axis {name} are not consistent between the event and background!")
                logger.error(f"        event      : {event_edges}")
                logger.error(f"        response : {response_edges}")
                raise ValueError

        if self._coordsys_conv_matrix is None:
            axes_cds = Axes([self._image_response.axes["Em"], \
                             self._image_response.axes["Phi"], \
                             self._image_response.axes["PsiChi"]])
        else:
            axes_cds = Axes([self._event.axes[0], \
                             self._image_response.axes["Em"], \
                             self._image_response.axes["Phi"], \
                             self._image_response.axes["PsiChi"]])
        
        self._event = Histogram(axes_cds, unit = self._event.unit, contents = self._event.contents)

        for key in self._bkg_models:
            bkg_model = self._bkg_models[key]
            self._bkg_models[key] = Histogram(axes_cds, unit = bkg_model.unit, contents = bkg_model.contents)
            del bkg_model

        logger.info(f"The axes in the event and background files are redefined. Now they are consistent with those of the response file.")

        return True

    def _load_full_detector_response_on_memory(self, full_detector_response, is_miniDC2_format):
        """
        Load a response file on the computer memory.
        """

        axes_image_response = [full_detector_response.axes["NuLambda"], full_detector_response.axes["Ei"],
                               full_detector_response.axes["Em"], full_detector_response.axes["Phi"], full_detector_response.axes["PsiChi"]]

        self._image_response = Histogram(axes_image_response, unit = full_detector_response.unit)

        nside = full_detector_response.axes["NuLambda"].nside
        npix = full_detector_response.axes["NuLambda"].npix 
    
        if is_miniDC2_format:
            for ipix in tqdm(range(npix)):
                self._image_response[ipix] = np.sum(full_detector_response[ipix].to_dense(), axis = (4,5)) #Ei, Em, Phi, ChiPsi
        else:
            contents = full_detector_response._file['DRM']['CONTENTS'][:]
            self._image_response[:] = contents * full_detector_response.unit

    def _calc_exposure_map(self):
        """
        Calculate exposure_map, which is an intermidiate matrix used in RL algorithm.
        """

        logger.info("Calculating an exposure map...")
        
        if self._coordsys_conv_matrix is None:
            self._exposure_map = Histogram(self._model_axes, unit = self._image_response.unit * u.sr)
            self._exposure_map[:] = np.sum(self._image_response.contents, axis = (2,3,4)) * self.model_axes['lb'].pixarea()
        else:
            self._exposure_map = Histogram(self._model_axes, unit = self._image_response.unit * self._coordsys_conv_matrix.unit * u.sr)
            self._exposure_map[:] = np.tensordot(np.sum(self._coordsys_conv_matrix, axis = (0)), 
                                                 np.sum(self._image_response, axis = (2,3,4)),
                                                 axes = ([1], [0]) ) * self._image_response.unit * self._coordsys_conv_matrix.unit * self.model_axes['lb'].pixarea()
            # [Time/ScAtt, lb, NuLambda] -> [lb, NuLambda]
            # [NuLambda, Ei, Em, Phi, PsiChi] -> [NuLambda, Ei]
            # [lb, NuLambda] x [NuLambda, Ei] -> [lb, Ei]

        logger.info("Finished...")

    def calc_expectation(self, model, dict_bkg_norm = None, almost_zero = 1e-12):
        """
        Calculate expected counts from a given model.

        Parameters
        ----------
        model : :py:class:`cosipy.image_deconvolution.AllSkyImageModel`
            Model map
        dict_bkg_norm : dict, default None
            background normalization for each background model, e.g, {'albedo': 0.95, 'activation': 1.05}
        almost_zero : float, default 1e-12
            In order to avoid zero components in extended count histogram, a tiny offset is introduced.
            It should be small enough not to effect statistics.

        Returns
        -------
        :py:class:`histpy.Histogram`
            Expected count histogram

        Notes
        -----
        This method should be implemented in a more general class, for example, extended source response class in the future.
        """
        # Currenly (2024-01-12) this method can work for both local coordinate CDS and in galactic coordinate CDS.
        # This is just because in DC2 the rotate response for galactic coordinate CDS does not have an axis for time/scatt binning.
        # However it is likely that it will have such an axis in the future in order to consider background variability depending on time and pointign direction etc.
        # Then, the implementation here will not work. Thus, keep in mind that we need to modify it once the response format is fixed.

        expectation = Histogram(self.data_axes)
        
        if self._coordsys_conv_matrix is None:
            expectation[:] = np.tensordot( model.contents, self._image_response.contents, axes = ([0,1],[0,1])) * model.axes['lb'].pixarea()
            # ['lb', 'Ei'] x [NuLambda(lb), Ei, Em, Phi, PsiChi] -> [Em, Phi, PsiChi]
        else:
            map_rotated = np.tensordot(self._coordsys_conv_matrix.contents, model.contents, axes = ([1], [0])) 
            # ['Time/ScAtt', 'lb', 'NuLambda'] x ['lb', 'Ei'] -> [Time/ScAtt, NuLambda, Ei]
            map_rotated *= self._coordsys_conv_matrix.unit * model.unit
            map_rotated *= model.axes['lb'].pixarea()
            # data.coordsys_conv_matrix.contents is sparse, so the unit should be restored.
            # the unit of map_rotated is 1/cm2 ( = s * 1/cm2/s/sr * sr)
            expectation[:] = np.tensordot( map_rotated, self._image_response.contents, axes = ([1,2], [0,1]))
            # [Time/ScAtt, NuLambda, Ei] x [NuLambda, Ei, Em, Phi, PsiChi] -> [Time/ScAtt, Em, Phi, PsiChi]

        if dict_bkg_norm is not None: 
            for key in self.keys_bkg_models():
                expectation += self.bkg_model(key) * dict_bkg_norm[key]
        expectation += almost_zero
        
        return expectation

    def calc_T_product(self, dataspace_histogram):
        """
        Calculate the product of the input histogram with the transonse matrix of the response function.
        Let R_{ij}, H_{i} be the response matrix and dataspace_histogram, respectively.
        Note that i is the index for the data space, and j is for the model space.
        In this method, \sum_{j} H{i} R_{ij}, namely, R^{T} H is calculated.

        Parameters
        ----------
        dataspace_histogram: :py:class:`histpy.Histogram`
            Its axes must be the same as self.data_axes

        Returns
        -------
        :py:class:`histpy.Histogram`
            The product with self.model_axes
        """
        # TODO: currently, dataspace_histogram is assumed to be a dense.

        hist_unit = self.exposure_map.unit
        if dataspace_histogram.unit is not None:
            hist_unit *= dataspace_histogram.unit

        hist = Histogram(self.model_axes, unit = hist_unit)

        if self._coordsys_conv_matrix is None:
            hist[:] = np.tensordot(dataspace_histogram.contents, self._image_response.contents, axes = ([0,1,2], [2,3,4])) * self.model_axes['lb'].pixarea()
            # [Em, Phi, PsiChi] x [NuLambda (lb), Ei, Em, Phi, PsiChi] -> [NuLambda (lb), Ei]
        else:
            _ = np.tensordot(dataspace_histogram.contents, self._image_response.contents, axes = ([1,2,3], [2,3,4])) 
            # [Time/ScAtt, Em, Phi, PsiChi] x [NuLambda, Ei, Em, Phi, PsiChi] -> [Time/ScAtt, NuLambda, Ei]

            hist[:] = np.tensordot(self._coordsys_conv_matrix.contents, _, axes = ([0,2], [0,1])) \
                        * _.unit * self._coordsys_conv_matrix.unit * self.model_axes['lb'].pixarea()
            # [Time/ScAtt, lb, NuLambda] x [Time/ScAtt, NuLambda, Ei] -> [lb, Ei]
            # note that coordsys_conv_matrix is the sparse, so the unit should be recovered.

        return hist

    def calc_bkg_model_product(self, key, dataspace_histogram):
        """
        Calculate the product of the input histogram with the background model.
        Let B_{i}, H_{i} be the background model and dataspace_histogram, respectively.
        In this method, \sum_{i} B_{i} H_{i} is calculated.

        Parameters
        ----------
        key: str
            Background model name
        dataspace_histogram: :py:class:`histpy.Histogram`
            its axes must be the same as self.data_axes

        Returns
        -------
        flaot
        """
        # TODO: currently, dataspace_histogram is assumed to be a dense.

        if self._coordsys_conv_matrix is None:

            return np.tensordot(dataspace_histogram.contents, self.bkg_model(key).contents, axes = ([0,1,2], [0,1,2]))

        return np.tensordot(dataspace_histogram.contents, self.bkg_model(key).contents, axes = ([0,1,2,3], [0,1,2,3]))

    def calc_loglikelihood(self, expectation):
        """
        Calculate log-likelihood from given expected counts or model/expectation.

        Parameters
        ----------
        expectation : :py:class:`histpy.Histogram`
            Expected count histogram.

        Returns
        -------
        float
            Log-likelood
        """
        loglikelood = np.sum( self.event * np.log(expectation) ) - np.sum(expectation)

        return loglikelood



import os
from awscli.clidriver import create_clidriver
from pathlib import Path

def fetch_wasabi_file(file,
                      output = None,
                      override = False,
                      bucket = 'cosi-pipeline-public',
                      endpoint = 'https://s3.us-west-1.wasabisys.com',
                      access_key = 'GBAL6XATQZNRV3GFH9Y4',
                      secret_key = 'GToOczY5hGX3sketNO2fUwiq4DJoewzIgvTCHoOv'):
    """
    Download a file from COSI's Wasabi acccount.

    Parameters
    ----------
    file : str
        Full path to file in Wasabi
    output : str,  optional
        Full path to the downloaded file in the local system. By default it will use 
        the current durectory and the same file name as the input file.
    bucket : str, optional
        Passed to aws --bucket option
    endpoint : str, optional
        Passed to aws --endpoint-url option
    access_key : str, optional
        AWS_ACCESS_KEY_ID
    secret_key : str, optional
        AWS_SECRET_ACCESS_KEY
    """
    
    if output is None:
        output = file.split('/')[-1]

    output = Path(output)    
        
    if output.exists() and not override:
        raise RuntimeError(f"File {output} already exists.")

    cli = create_clidriver()

    cli.session.set_credentials(access_key, secret_key)
    command = ['s3api', 'get-object',
               '--bucket', bucket,
               '--key', file,
               '--endpoint-url', endpoint,
               str(output)]

    cli.main(command)




# Imports:
import os
from astropy.coordinates import SkyCoord
from astropy import units as u
from cosipy.response import FullDetectorResponse, DetectorResponse
from cosipy import BinnedData
from mhealpy import HealpixMap, HealpixBase
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import norm
import numpy.ma as ma
from tqdm import tqdm
import logging
logger = logging.getLogger(__name__)

class ContinuumEstimation:
    
    def calc_psr(self, sc_orientation, detector_response, coord, nside=16):

        """Calculates point source response (PSR) in Galactic coordinates.
        
        Parameters
        ----------
        ori_file : str
            Full path to orienation file.
        sc_orientation : cosipy.spacecraftfile.SpacecraftFile
            Spacecraft orientation object.  
        detector_response : str
            Full path to detector response file.
        coord : astropy.coordinates.SkyCoord
            The coordinates of the target object. 
        nside : int, optional
            nside of scatt map (default is 16). 
        
        Returns
        -------
        :py:class:`PointSourceResponse` 
        """

        # Detector response:
        dr = detector_response

        # Scatt map:
        scatt_map = sc_orientation.get_scatt_map(coord, nside = nside, coordsys = 'galactic')

        # Calculate PSR:
        with FullDetectorResponse.open(dr) as response:
            psr = response.get_point_source_response(coord = coord, scatt_map = scatt_map)

        return psr

    def load_psr_from_file(self, psr_file):

        """Loads point source response from h5 file.
            
        Parameters
        ----------
        psr_file : str
            Full path to precomputed response file (.h5 file).
        """

        logger.info("...loading the pre-computed point source response ...")
        psr = DetectorResponse.open(psr_file)
        logger.info("--> done")

        return psr

    def load_full_data(self, data_file, data_yaml):

        """Loads binned data to be used as a template for the background estimate.
        
        Parameters
        ----------
        data_file : str
            Full path to binned data (must be .h5 file). 
        data_yaml : str
            Full path to the dataIO yaml file used for binning the data. 

        Notes
        -----
        In practice, the data file used for estimating the background 
        should be the full dataset.  
        
        The full data binning needs to match the PSR. 
        """
        
        self.full_data = BinnedData(data_yaml)
        self.full_data.load_binned_data_from_hdf5(data_file)

        return

    def mask_from_cumdist(self, psichi_map, containment, make_plots=False):

        """
        Determines masked pixels from cumulative distribution of
        the point source response.
        
        Parameters
        ----------
        psichi_map : histpy:Histogram
            Point source response projected onto psichi. This can be 
            either a slice of Em and Phi, or the full projection. Note
            that psichi is a HealpixMap axis in histpy. 
        containment : float
            The percentage (non-inclusive) of the cumulative distribution 
            to use for the mask, i.e. all pixels that fall below this value 
            in the cumulative distribution will be masked. 
        make_plots : bool
            Option to plot cumulative distribution. 

        Returns
        -------
        sorted_indices : array
            Indices of sorted psichi array.
        arm_mask : array
            Boolean array specifying pixels in the psichi map that will be masked. 

        Note
        ----
        The cumulative distribution is an estimate of the angular
        resolution measure (ARM), which is a measure of the PSF
        for Compton imaging. 
        """

        # Get healpix map:
        h = psichi_map
        m = HealpixMap(base = HealpixBase(npix = h.nbins), data = h.contents)

        # Sort data in descending order:
        sorted_data = np.sort(m)[::-1]

        # Calculte the cummulative distribution
        cumdist = np.cumsum(sorted_data) / sum(sorted_data)

        # Get indices of sorted array
        sorted_indices = np.argsort(h.contents.value)[::-1]

        # Define mask based on fraction of total exposure (i.e. counts):
        arm_mask = cumdist >= containment
        arm_mask = ~arm_mask

        # Plot cummulative distribution and corresponding masks:
        if make_plots == True:
            plt.plot(cumdist)
            plt.title("Cumulative Distribution")
            plt.xlabel("Pixel")
            plt.ylabel("Fraction of Counts")
            plt.savefig("cumdist.png")
            plt.show()
            plt.close()
        
        return sorted_indices, arm_mask

    def simple_inpainting(self, m_data, sorted_indices, arm_mask):

        """Highly simplistic method for inpainting masked region in CDS.

        This method relies on the input healpix map having a ring 
        ordering. For each masked pixel, it searches to the left (i.e.
        lower pixel numbers) until reaching the first non-zero pixel. 
        It then search to the right (i.e. higher pixel numbers) until
        again finding the first non-zero pixel. The mean of the two 
        values is used for filling in the masked pixel. 
        
        Parameters
        ----------
        m_data : array-like
            HealpixMap object, containing projection of PSR onto psichi.
        sorted_indices : array
            Indices of sorted psichi array.
        arm_mask : array
            Boolean array specifying pixels in the psichi map that will be masked. 

        Returns
        -------
        interp_list : array
            Values for the inpainting, corresponding to the masked pixels. 
        """

        # Get mean of masked data for edge cases (simple solution for now):
        # CK: It would be better if this were at least the mean of an 
        # np masked array object, but a better method is anyways needed.
        masked_mean = np.mean(m_data)

        # Get interpolation values:
        interp_list_low = []
        interp_list_high = []
        for i in range(0,len(sorted_indices[arm_mask])):
            
            this_index = sorted_indices[arm_mask][i]
            
            # Search left:
            k = 1
            search_left = True
            while search_left == True:
                
                if this_index-k < 0:
                    logger.info("Edge case!")
                    interp_list_low.append(masked_mean)
                    search_left = False
                    break
                    
                next_value = m_data[this_index-k]
                if next_value == 0:
                    k += 1
                if next_value != 0:
                    interp_list_low.append(next_value)
                    search_left = False
           
            # Search right:
            j = 1
            search_right = True
            while search_right == True:
               
                if this_index+j >= self.psr.axes['PsiChi'].nbins-1:
                    logger.info("Edge case!")
                    interp_list_high.append(masked_mean)
                    search_right = False
                    break
                
                next_value = m_data[this_index+j]
                if next_value == 0:
                    j += 1
                if next_value != 0:
                    interp_list_high.append(next_value)
                    search_right = False
            
        interp_list_low = np.array(interp_list_low)
        interp_list_high = np.array(interp_list_high)
        interp_list = (interp_list_low + interp_list_high) / 2.0 
    
        return interp_list

    def continuum_bg_estimation(self, data_file, data_yaml, psr, \
            containment=0.4, make_plots=False,\
            e_loop="default", s_loop="default"):

        """Estimates continuum background.
        
        Parameters
        ----------
        data_file : str
            Full path to binned data (must be .h5 file). 
        data_yaml : str
            Full path to the dataIO yaml file used for binning the data.  
        psr : py:class:`PointSourceResponse`
            Point source response object. 
        containment : float, optional
            The percentage (non-inclusive) of the cumulative distribution 
            to use for the mask, i.e. all pixels that fall below this value 
            in the cumulative distribution will be masked. Default is 0.4. 
        make_plots : bool, optional
            Option to make some plots of the data, response, and masks. 
            Default is False.
        e_loop : tuple, optional
            Option to pass tuple specifying which energy range to 
            loop over. This must coincide with the energy bins. The default
            is all bins.
        s_loop : tuple, optional
            Option to pass tuple specifying which Phi anlge range to
            loop over. This must coincide with the Phi  bins. The default
            is all bins.
        
        Returns
        -------
        estimated_bg : histpy:Histogram
            Estimated background as histpy object. 
        """

        # Define psr attribute:
        self.psr = psr

        # Load data to be used for BG estimation:
        self.load_full_data(data_file,data_yaml)
        estimated_bg = self.full_data.binned_data.project('Em', 'Phi', 'PsiChi')

        # Defaults for energy and scattering angle loops:
        if e_loop == "default":
            e_loop = (0,len(self.psr.axes['Em'].centers))
        if s_loop == "default":
            s_loop = (0,len(self.psr.axes['Phi'].centers))

        # Progress bar:
        e_tot = e_loop[1] - e_loop[0]
        s_tot = s_loop[1] - s_loop[0]
        num_lines = e_tot*s_tot
        pbar = tqdm(total=num_lines)

        # Loop through all bins of energy and phi:
        for E in range(e_loop[0],e_loop[1]):
            for s in range(s_loop[0],s_loop[1]):
                
                pbar.update(1) # update progress bar
                logger.info("Bin %s %s" %(str(E),str(s)))

                # Get PSR slice:
                h = self.psr.slice[{'Em':E, 'Phi':s}].project('PsiChi')

                # Get mask:
                sorted_indices, arm_mask = self.mask_from_cumdist(h, containment, make_plots=make_plots)       

                # Mask data:
                h_data = self.full_data.binned_data.project('Em', 'Phi', 'PsiChi').slice[{'Em':E, 'Phi':s}].project('PsiChi')
                m_data = HealpixMap(base = HealpixBase(npix = h_data.nbins), data = h_data.contents.todense())
                m_data[sorted_indices[arm_mask]] = 0

                # Skip this iteration if map is all zeros:
                if len(m_data[m_data[:] > 0]) == 0:
                    logger.info("All zeros and so skipping iteration!")
                    continue

                # Get interpolated values:
                interp_list = self.simple_inpainting(m_data, sorted_indices, arm_mask)

                # Update estimated BG:
                for p in range(len(sorted_indices[arm_mask])):
                    estimated_bg[E,s,sorted_indices[arm_mask][p]] = interp_list[p]

                # Option to make some plots:
                if make_plots == True:
                    
                    # Plot true response:
                    m_dummy = HealpixMap(base = HealpixBase(npix = h.nbins), data = h.contents)
                    plot,ax = m_dummy.plot('mollview')
                    plt.title("True Response")
                    plt.show()
                    plt.close()

                    # Plot masked response:
                    m_dummy[sorted_indices[arm_mask]] = 0
                    plot,ax = m_dummy.plot('mollview')
                    plt.title("Masked Response")
                    plt.show()
                    plt.close()

                    # Plot true data:
                    m_data_dummy = HealpixMap(base = HealpixBase(npix = h_data.nbins), data = h_data.contents.todense())
                    plot,ax = m_data_dummy.plot('mollview')
                    plt.title("True Data")
                    plt.show()
                    plt.close()

                    # Plot masked data:
                    plot,ax = m_data.plot('mollview')
                    plt.title("Masked Data")
                    plt.show()
                    plt.close()

                    # Plot masked data with interpolated values:
                    m_data[sorted_indices[arm_mask]] = interp_list
                    plot,ax = m_data.plot('mollview')
                    plt.title("Interpolated Data (Estimated BG)")
                    plt.show()
                    plt.close()
      
        # Close progress bar:
        pbar.close()

        return estimated_bg



from abc import ABC, abstractmethod

class ImageDeconvolutionDataInterfaceBase(ABC):
    """
    A base class for managing data for image analysis, i.e., 
    event data, background models, response, coordsys conversion matrix etc.
    Subclasses must override these attributes and methods.

    Attributes:
    - self._event 
        A binned histogram of events. It is an instance of histpy.Histogram.
        Its axes must be the same of self._data_axes.
    - self._bkg_models
        A dictionary of binned histograms of background models.
        It is a dictionary of histpy.Histogram with keys of names of background models.
        Their axes must be the same of self._model_axes.
    - self._summed_bkg_models
        A dictionary of summed values of the background model histograms.
    - self._exposure_map 
        A binned histogram of the exposures at each pixel in the model space.
        It is an instance of histpy.Histogram.
        Its axes must be the same of self._model_axes.
    - self._model_axes 
        Axes for the data space. It is an instance of histpy.Axes.
    - self._data_axes 
        Axes for the model space. It is an instance of histpy.Axes.
    
    Methods:
    - keys_bkg_models()
        It returns a list of names of background models.
    - bkg_model(key)
        It returns a binned histogram of a background model with the given key.
    - summed_bkg_model(key)
        It returns the summed value of the background histogram with the given key.
    - calc_expectation(model)
        It returns a histogram of expected counts from the given model.
    - calc_T_product(dataspace_histogram)
        It returns the product of the input histogram with the transonse matrix of the response function.
    - calc_bkg_model_product(key, dataspace_histogram)
        It returns the product of an input histogram with a background model with the given key.
    - calc_likelihood(expectation)
        It returns the log-likelihood with the given histogram of expected counts.

    The basic idea of this class is to separate the data structure
    from the development of the image deconvolution algorithm.
    
    When the image deconvolution is performed, the deconvolution algorithm will look at only the above 
    attributes and methods, and will not care about the actual response matrix and how it actually calculates
    expected counts or the product using the transpose matrix of the response function.
    """

    def __init__(self, name = None):
        self._name = name

        # must assign data to them somewhere
        self._event = None # histpy.Histogram
        self._bkg_models = {} # a dictionary of histpy.Histogram
        self._summed_bkg_models = {} # a dictionary of float
        self._exposure_map = None # histpy.Histogram
        self._model_axes = None # histpy.Axes
        self._data_axes = None # histpy.Axes

    @property
    def name(self):
        return self._name

    @property
    def event(self):
        return self._event

    @property
    def exposure_map(self):
        return self._exposure_map

    @property
    def model_axes(self):
        return self._model_axes

    @property
    def data_axes(self):
        return self._data_axes

    def keys_bkg_models(self):
        return list(self._bkg_models.keys())

    def bkg_model(self, key):
        return self._bkg_models[key]

    def summed_bkg_model(self, key):
        return self._summed_bkg_models[key]

    @abstractmethod
    def calc_expectation(self, model, dict_bkg_norm = None, almost_zero = 1e-12):
        """
        Calculate expected counts from a given model map.

        Parameters
        ----------
        model : :py:class:`cosipy.image_deconvolution.ModelBase` or its subclass
            Model
        dict_bkg_norm : dict, default None
            background normalization for each background model, e.g, {'albedo': 0.95, 'activation': 1.05}
        almost_zero : float, default 1e-12
            In order to avoid zero components in extended count histogram, a tiny offset is introduced.
            It should be small enough not to effect statistics.

        Returns
        -------
        :py:class:`histpy.Histogram`
            Expected count histogram
        """
        raise NotImplementedError

    @abstractmethod
    def calc_T_product(self, dataspace_histogram):
        """
        Calculate the product of the input histogram with the transonse matrix of the response function.
        Let R_{ij}, H_{i} be the response matrix and dataspace_histogram, respectively.
        Note that i is the index for the data space, and j is for the model space.
        In this method, \sum_{j} H{i} R_{ij}, namely, R^{T} H is calculated.

        Parameters
        ----------
        dataspace_histogram: :py:class:`histpy.Histogram`
            Its axes must be the same as self.data_axes

        Returns
        -------
        :py:class:`histpy.Histogram`
            The product with self.model_axes
        """
        raise NotImplementedError
    
    @abstractmethod
    def calc_bkg_model_product(self, key, dataspace_histogram):
        """
        Calculate the product of the input histogram with the background model.
        Let B_{i}, H_{i} be the background model and dataspace_histogram, respectively.
        In this method, \sum_{i} B_{i} H_{i} is calculated.

        Parameters
        ----------
        key: str
            Background model name
        dataspace_histogram: :py:class:`histpy.Histogram`
            its axes must be the same as self.data_axes

        Returns
        -------
        float
        """
        raise NotImplementedError
    
    @abstractmethod
    def calc_loglikelihood(self, expectation):
        """
        Calculate log-likelihood from given expected counts or model/expectation.

        Parameters
        ----------
        expectation : :py:class:`histpy.Histogram`
            Expected count histogram.

        Returns
        -------
        float
            Log-likelood
        """
        raise NotImplementedError



import numpy as np
from astropy.coordinates import Angle
import astropy.units as u
from astropy.stats import poisson_conf_interval
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from cosipy.polarization.polarization_angle import PolarizationAngle
from cosipy.polarization.conventions import MEGAlibRelativeX, IAUPolarizationConvention
from cosipy.response import FullDetectorResponse
from scoords import SpacecraftFrame

def calculate_uncertainties(counts):
    """
    Calculate the Poisson/Gaussian uncertainties for a list of binned counts.
        
    Parameters
    ----------
    counts : list
        List of counts in each bin

    Returns
    -------
    uncertainties : np.ndarray
        Lower & upper uncertainties for each bin
    """
    
    uncertainties_low = []
    uncertainties_high = []
    for i in range(len(counts)):
        if counts[i] <= 5:
            poisson_uncertainty = poisson_conf_interval(counts[i], interval="frequentist-confidence", sigma=1)
            uncertainties_low.append(counts[i] - poisson_uncertainty[0])
            uncertainties_high.append(poisson_uncertainty[1] - counts[i])
        else:
            gaussian_uncertainty = np.sqrt(counts[i])
            uncertainties_low.append(gaussian_uncertainty)
            uncertainties_high.append(gaussian_uncertainty)

    uncertainties = np.array([uncertainties_low, uncertainties_high])
        
    return uncertainties

class PolarizationASAD():
    """
    Azimuthal scattering angle distribution (ASAD) method to fit polarization.

    Parameters
    ----------
    source_vector : astropy.coordinates.sky_coordinate.SkyCoord
        Source direction
    source_spectrum : astromodels.functions.functions_1D
        Spectrum of source
    response_file : str or pathlib.Path
        Path to detector response
    sc_orientation : cosipy.spacecraftfile.SpacecraftFile.SpacecraftFile
        Spacecraft orientation
    fit_convention : cosipy.polarization.conventions.PolarizationConvention, optional
        Polarization reference convention to use for fit. Default is RelativeX
    """

    def __init__(self, source_vector, source_spectrum, response_file, sc_orientation, fit_convention=None):
        
        if fit_convention == None:
            self._convention = MEGAlibRelativeX(attitude=source_vector.attitude)
        else:
            self._convention = fit_convention

        reference_vector = self._convention.get_basis(source_vector)[0] #px

        if isinstance(source_vector.frame, SpacecraftFrame):
            self._source_vector = source_vector
        else:
            self._source_vector = source_vector.transform_to(SpacecraftFrame(attitude=source_vector.attitude))
        
        if isinstance(reference_vector.frame, SpacecraftFrame):
            self._reference_vector = reference_vector
        else:
            self._reference_vector = reference_vector.transform_to(SpacecraftFrame(attitude=source_vector.attitude))

        self._source_vector_cartesian = [self._source_vector.cartesian.x.value,
                                         self._source_vector.cartesian.y.value, 
                                         self._source_vector.cartesian.z.value]
        self._reference_vector_cartesian = [self._reference_vector.cartesian.x.value, 
                                            self._reference_vector.cartesian.y.value, 
                                            self._reference_vector.cartesian.z.value]
        
        self._expectation, self._azimuthal_angle_bins = self.convolve_spectrum(source_spectrum, response_file, sc_orientation)

        self._energy_range = [min(self.response.axes['Em'].edges.value), max(self.response.axes['Em'].edges.value)]
        
    def convolve_spectrum(self, spectrum, response_file, sc_orientation):
        """
        Convolve source spectrum with response and calculate azimuthal scattering angle bins.

        Parameters
        ----------
        response_file : str or pathlib.Path
            Path to detector response
        sc_orientation : cosipy.spacecraftfile.SpacecraftFile.SpacecraftFile
            Spacecraft orientation

        Returns
        -------
        expectation : cosipy.response.PointSourceResponse.PointSourceResponse
            Expected counts in each bin of Compton data space
        azimuthal_angle_bins : list
            Centers of azimuthal scattering angle bins calculated from PsiChi bins in response
        """

        self.response = FullDetectorResponse.open(response_file)

        target_in_sc_frame = sc_orientation.get_target_in_sc_frame(target_name='source', target_coord=self._source_vector.transform_to('galactic'))
        dwell_time_map = sc_orientation.get_dwell_map(response=response_file, src_path=target_in_sc_frame)

        psr = self.response.get_point_source_response(exposure_map=dwell_time_map, coord=self._source_vector.transform_to('galactic'))

        expectation = psr.get_expectation(spectrum)

        azimuthal_angle_bins = []

        for i in range(expectation.axes['PsiChi'].nbins):
            azimuthal_angle = self.calculate_azimuthal_scattering_angle(expectation.project(['PsiChi']).axes['PsiChi'].pix2ang(i)[0], expectation.project(['PsiChi']).axes['PsiChi'].pix2ang(i)[1])
            azimuthal_angle_bins.append(azimuthal_angle)

        return expectation, azimuthal_angle_bins

    def calculate_azimuthal_scattering_angle(self, psi, chi):
        """
        Calculate the azimuthal scattering angle of a scattered photon.
        
        Parameters
        ----------
        psi : float
            Polar angle (radians) of scattered photon in local coordinates
        chi : float
            Azimuthal angle (radians) of scattered photon in local coordinates

        Returns
        -------
        azimuthal_angle : astropy.coordinates.Angle
            Azimuthal scattering angle defined with respect to given reference vector
        """
        
        # Convert scattered photon vector from spherical to Cartesian coordinates
        scattered_photon_vector = [np.sin(psi) * np.cos(chi), np.sin(psi) * np.sin(chi), np.cos(psi)]

        # Project scattered photon vector onto plane perpendicular to source direction
        d = np.dot(scattered_photon_vector, self._source_vector_cartesian) / np.dot(self._source_vector_cartesian, self._source_vector_cartesian)
        projection = [scattered_photon_vector[0] - (d * self._source_vector_cartesian[0]), 
                      scattered_photon_vector[1] - (d * self._source_vector_cartesian[1]), 
                      scattered_photon_vector[2] - (d * self._source_vector_cartesian[2])]

        # Calculate angle between scattered photon vector & reference vector on plane perpendicular to source direction
        cross_product = np.cross(projection, self._reference_vector_cartesian)
        if np.dot(self._source_vector_cartesian, cross_product) < 0:
            sign = -1
        else:
            sign = 1
        normalization = np.sqrt(np.dot(projection, projection)) * np.sqrt(np.dot(self._reference_vector_cartesian, self._reference_vector_cartesian))
    
        azimuthal_angle = Angle(sign * np.arccos(np.dot(projection, self._reference_vector_cartesian) / normalization), unit=u.rad)
    
        return azimuthal_angle

    def calculate_azimuthal_scattering_angles(self, unbinned_data):
        """
        Calculate the azimuthal scattering angles for all events in a dataset.
        
        Parameters
        ----------
        unbinned_data : dict
            Unbinned data including polar and azimuthal angles (radians) of scattered photon in local coordinates

        Returns
        -------
        azimuthal_angles : list
            Azimuthal scattering angles. Each angle must be an astropy.coordinates.Angle object
        """

        azimuthal_angles = []

        for i in range(len(unbinned_data['Psi local'])):
            if unbinned_data['Energies'][i] >= self._energy_range[0] and unbinned_data['Energies'][i] <= self._energy_range[1]:
                azimuthal_angle = self.calculate_azimuthal_scattering_angle(unbinned_data['Psi local'][i], unbinned_data['Chi local'][i])
                azimuthal_angles.append(azimuthal_angle)

        return azimuthal_angles

    def create_asad(self, azimuthal_angles, bins=20):
        """
        Create ASAD and calculate uncertainties.
        
        Parameters
        ----------
        azimuthal_angles : list
            Azimuthal scattering angles (radians)
        bins : int or np.array, optional
            Number of azimuthal scattering angle bins if int or edges of azimuthal scattering angle bins if np.array (radians)

        Returns
        -------
        asad : dict
            Counts and Gaussian/Poisson errors in each azimuthal scattering angle bin
        """

        if isinstance(bins, int):
            bin_edges = Angle(np.linspace(-np.pi, np.pi, bins), unit=u.rad)
        else:
            bin_edges = bins

        counts, edges = np.histogram(azimuthal_angles, bins=bin_edges)
        self._bin_edges = edges
        self._bins = []
        for i in range(len(self._bin_edges) - 1):
            self._bins.append((self._bin_edges[i] + self._bin_edges[i+1]) / 2)
        errors = calculate_uncertainties(counts)

        asad = {'counts': counts, 'uncertainties': errors}

        return asad
    
    def create_unpolarized_asad(self, bins=None):
        """
        Calculate the azimuthal scattering angles for all bins.
        
        Parameters
        ----------
        bins : int or np.array, optional
            Number of azimuthal scattering angle bins if int or edges of azimuthal scattering angle bins if np.array (radians)

        Returns
        -------
        asad : dict
            Counts and Gaussian/Poisson errors in each azimuthal scattering angle bin
        """

        if not bins == None:
            if isinstance(bins, int):
                bin_edges = Angle(np.linspace(-np.pi, np.pi, bins), unit=u.rad)
            else:
                bin_edges = bins
        else:
            bin_edges = self._bin_edges
        
        unpolarized_asad = []

        for i in range(len(bin_edges)-1):
            counts = 0
            for j in range(self._expectation.project(['PsiChi']).nbins):
                if self._azimuthal_angle_bins[j] >= bin_edges[i] and self._azimuthal_angle_bins[j] < bin_edges[i+1]:
                    counts += self._expectation.project(['PsiChi'])[j]
            unpolarized_asad.append(counts)

        asad = {'counts': unpolarized_asad, 'uncertainties': calculate_uncertainties(unpolarized_asad)}

        return asad
    
    def create_polarized_asads(self, bins=None):
        """
        Calculate the azimuthal scattering angles for all bins.
        
        Parameters
        ----------
        bins : int or np.array, optional
            Number of azimuthal scattering angle bins if int or edges of azimuthal scattering angle bins if np.array (radians)

        Returns
        -------
        asads : dict
            Counts and Gaussian/Poisson errors in each azimuthal scattering angle bin for each polarization angle bin
        """

        if not bins == None:
            if isinstance(bins, int):
                bin_edges = Angle(np.linspace(-np.pi, np.pi, bins), unit=u.rad)
            else:
                bin_edges = bins
        else:
            bin_edges = self._bin_edges
        
        polarized_asads = {}
        polarized_asad_uncertainties = {}
        for k in range(self._expectation.axes['Pol'].nbins):
            polarized_asads[k] = []
            for i in range(len(bin_edges)-1):
                counts = 0
                for j in range(self._expectation.project(['PsiChi']).nbins):
                    if self._azimuthal_angle_bins[j] >= bin_edges[i] and self._azimuthal_angle_bins[j] < bin_edges[i+1]:
                        counts += self._expectation.slice[{'Pol':slice(k,k+1)}].project(['PsiChi'])[j]
                polarized_asads[k].append(counts)
            polarized_asad_uncertainties[k] = calculate_uncertainties(polarized_asads[k])

        asads = {'counts': polarized_asads, 'uncertainties': polarized_asad_uncertainties}

        return asads

    def asad_sinusoid(self, x, a, b, c):
        """
        Sinusoid to fit to ASAD.
        
        Parameters
        ----------
        x : float
            Azimuthal scattering angle (radians)
        a : float
            First parameter
        b : float
            Second parameter
        c : float
            Third parameter
            
        Returns
        -------
        asad_function : float
            Y-value of ASAD
        """

        asad_function = a - (b * np.cos(2 * (x - c)))
    
        return asad_function
    
    def fit_asad(self, counts, p0, bounds, sigma):
        """
        Fit the ASAD with a sinusoid.
        
        Parameters
        ----------
        counts : list
            Counts in each azimuthal scattering angle bin
        p0 : list or np.array
            Initial guess for parameter values
        bounds : 2-tuple of float, list, or np.array
            Lower & upper bounds on parameters
        sigma : float, list, or np.array
            Uncertainties in y data
            
        Returns
        -------
        popt : np.ndarray
            Fitted parameter values
        uncertainties : list
            Uncertainty on each parameter value
        """

        popt, pcov = curve_fit(self.asad_sinusoid, Angle(self._bins).rad, counts, p0=p0, bounds=bounds, sigma=sigma)
        uncertainties = [] 
        for i in range(len(pcov)):
            uncertainties.append(np.sqrt(pcov[i][i]))

        return popt, uncertainties
    
    def plot_asad(self, counts, error, title, coefficients=[]):
        """
        Plot the ASAD.
        
        Parameters
        ----------
        counts : list
            Counts in each azimuthal scattering angle bin
        error : np.ndarray
            Lower & upper uncertainties for each bin
        title : str
            Title of plot
        coefficients : list, optional
            Coefficients to plot fitted sinusoidal function
        """

        plt.scatter(Angle(self._bins).degree, counts)
        plt.errorbar(Angle(self._bins).degree, counts, yerr=error, linewidth=0, elinewidth=1)
        plt.title(title)
        plt.xlabel('Azimuthal Scattering Angle (degrees)')
        
        if len(coefficients) == 3:
            x = np.linspace(-np.pi, np.pi, 1000)
            y = []
            for item in x:
                y.append(self.asad_sinusoid(item, coefficients[0], coefficients[1], coefficients[2]))
            plt.plot(list(np.rad2deg(x)), y, color='green')

        plt.show()

    def correct_asad(self, data_asad, unpolarized_asad):
        """
        Correct the ASAD using the ASAD of an unpolarized source.
        
        Parameters
        ----------
        data_asad : dict
            Counts and uncertainties in each azimuthal scattering angle bin of data
        unpolarized_asad : dict
            Counts and uncertainties in each azimuthal scattering angle bin of unpolarized source
            
        Returns
        -------
        asad : dict
            Normalized counts and uncertainties in each azimuthal scattering angle bin
        """
    
        corrected = []
        for i in range(len(self._bins)):
            corrected.append(data_asad['counts'][i] / np.sum(data_asad['counts']) / unpolarized_asad['counts'][i] * np.sum(unpolarized_asad['counts']))
  
        errors_low = []
        errors_high = []
        for i in range(len(self._bins)):
            sigma_corrected_low = corrected[i] * np.sqrt(((data_asad['uncertainties'][0][i])/data_asad['counts'][i])**2 + ((unpolarized_asad['uncertainties'][0][i])/unpolarized_asad['counts'][i])**2)
            sigma_corrected_high = corrected[i] * np.sqrt(((data_asad['uncertainties'][1][i])/data_asad['counts'][i])**2 + ((unpolarized_asad['uncertainties'][1][i])/unpolarized_asad['counts'][i])**2)
            errors_low.append(sigma_corrected_low)
            errors_high.append(sigma_corrected_high)

        asad = {'counts': corrected, 'uncertainties': np.array([errors_low, errors_high])}

        return asad

    def calculate_mu(self, counts_corrected, p0=None, bounds=(-np.inf, np.inf), sigma=None):
        """
        Calculate the modulation (mu).
        
        Parameters
        ----------
        counts_corrected : list
            Counts in each azimuthal scattering angle bin
        p0 : list or np.array
            Initial guess for parameter values
        bounds : 2-tuple of float, list, or np.array
            Lower & upper bounds on parameters
        sigma : float, list, or np.array
            Uncertainties for each azimuthal scattering angle bin
            
        Returns
        -------
        modulation : dict
            Modulation and uncertainty of fitted sinusoid
        """

        if isinstance(sigma, np.ndarray) and len(sigma.shape) == 2:
            for i in range(len(sigma[0])):
                if sigma[0][i] != sigma[1][i]:
                    print('Warning: Uncertainty in at least one bin of ASAD is not Gaussian. Making error bars symmetric. Fit may not be accurate.')
                    break
            symmetric_sigma = []
            for i in range(len(sigma[0])):
                symmetric_sigma.append((sigma[0][i] + sigma[1][i]) / 2)
            sigma = symmetric_sigma

        parameter_values, uncertainties = self.fit_asad(counts_corrected, p0, bounds, sigma)
    
        mu = parameter_values[1] / parameter_values[0]
        mu_uncertainty = mu * np.sqrt((uncertainties[0]/parameter_values[0])**2 + (uncertainties[1]/parameter_values[1])**2)

        modulation = {'mu': mu, 'uncertainty': mu_uncertainty}

        print('Modulation:', round(mu, 3), '+/-', round(mu_uncertainty, 3))
    
        return modulation
    
    def constant(self, x, a):
        """
        Constant function to fit to mu_100 values.
        
        Parameters
        ----------
        x : float
            Mu_100
        a : float
            Parameter
            
        Returns
        -------
        a : float
            Constant value
        """

        return a
    
    def calculate_mu100(self, polarized_asads, unpolarized_asad):
        """
        Calculate the modulation (mu) of an 100% polarized source.
        
        Parameters
        ----------
        polarized_asads : list
            Counts and Gaussian/Poisson errors in each azimuthal scattering angle bin for each polarization angle bin for 100% polarized source
        unpolarized_asad : list or np.array
            Counts and Gaussian/Poisson errors in each azimuthal scattering angle bin for unpolarized source
            
        Returns
        -------
        mu_100 : dict
            Modulation of 100% polarized source and uncertainty of constant function fit to modulation in all polarization angle bins
        """

        mu_100_list = []
        mu_100_uncertainties = []
        for i in range(self._expectation.axes['Pol'].nbins):
            print('Polarization angle bin: ' + str(self._expectation.axes['Pol'].edges[i]) + ' to ' + str(self._expectation.axes['Pol'].edges[i+1]))
            asad_polarized = {'counts': polarized_asads['counts'][i], 'uncertainties': polarized_asads['uncertainties'][i]}
            asad_polarized_corrected = self.correct_asad(asad_polarized, unpolarized_asad)
            mu_100 = self.calculate_mu(asad_polarized_corrected['counts'], bounds=([0, 0, 0], [np.inf,np.inf,np.pi]), sigma=asad_polarized_corrected['uncertainties'])
            mu_100_list.append(mu_100['mu'])
            mu_100_uncertainties.append(mu_100['uncertainty'])
            self.plot_asad(asad_polarized_corrected['counts'], asad_polarized_corrected['uncertainties'], 'Corrected 100% Polarized ASAD', coefficients=self.fit(mu_100, asad_polarized_corrected['counts'], bounds=([0, 0, 0], [np.inf,np.inf,np.pi]), sigma=asad_polarized_corrected['uncertainties'])['best fit parameter values'])

        popt, pcov = curve_fit(self.constant, self._expectation.axes['Pol'].centers, mu_100_list, sigma=mu_100_uncertainties)
        mu_100 = {'mu': popt[0], 'uncertainty': pcov[0][0]}

        plt.scatter(self._expectation.axes['Pol'].centers, mu_100_list)
        plt.errorbar(self._expectation.axes['Pol'].centers, mu_100_list, yerr=mu_100_uncertainties, linewidth=0, elinewidth=1)
        plt.plot([0, 175], [mu_100['mu'], mu_100['mu']])
        plt.xlabel('Polarization Angle (degrees)')
        plt.ylabel('mu_100')
        plt.show()

        print('mu_100:', round(mu_100['mu'], 2))

        return mu_100

    def fit(self, mu_100, counts_corrected, p0=None, bounds=(-np.inf, np.inf), sigma=None):
        """
        Fit the polarization fraction and angle.
        
        Parameters
        ----------
        mu_100 : dict
            Modulation and uncertainty of a 100% polarized source
        counts_corrected : list
            Counts in each azimuthal scattering angle bin
        p0 : list or np.array
            Initial guess for parameter values
        bounds : 2-tuple of float, list, or np.array
            Lower & upper bounds on parameters
        sigma : float, list, or np.array
            Uncertainties for each azimuthal scattering angle bin
            
        Returns
        -------
        polarization : dict
            Polarization fraction, polarization angle, and best fit parameter values for fitted sinusoid, and associated uncertainties
        """

        if isinstance(sigma, np.ndarray) and len(sigma.shape) == 2:
            for i in range(len(sigma[0])):
                if sigma[0][i] != sigma[1][i]:
                    print('Warning: Uncertainty in at least one bin of ASAD is not Gaussian. Making error bars symmetric. Fit may not be accurate.')
                    break
            symmetric_sigma = []
            for i in range(len(sigma[0])):
                symmetric_sigma.append((sigma[0][i] + sigma[1][i]) / 2)
            sigma = symmetric_sigma

        parameter_values, uncertainties = self.fit_asad(counts_corrected, p0, bounds, sigma)
    
        polarization_fraction = parameter_values[1] / (parameter_values[0] * mu_100['mu'])
        polarization_fraction_uncertainty = polarization_fraction * np.sqrt((uncertainties[0]/parameter_values[0])**2 + (uncertainties[1]/parameter_values[1])**2 + (mu_100['uncertainty']/mu_100['mu'])**2)

        polarization_angle = Angle(parameter_values[2], unit=u.rad)
        polarization_angle.wrap_at(180 * u.deg, inplace=True)
        if polarization_angle.degree < 0:
            polarization_angle += Angle(180, unit=u.deg)
        polarization_angle = PolarizationAngle(polarization_angle, self._source_vector, convention=self._convention).transform_to(IAUPolarizationConvention())
        polarization_angle_uncertainty = Angle(uncertainties[2], unit=u.rad)

        polarization = {'fraction': polarization_fraction, 'angle': polarization_angle, 'fraction uncertainty': polarization_fraction_uncertainty, 'angle uncertainty': polarization_angle_uncertainty, 'best fit parameter values': parameter_values, 'best fit parameter uncertainties': uncertainties}
    
        print('Best fit polarization fraction:', round(polarization_fraction, 3), '+/-', round(polarization_fraction_uncertainty, 3))
        print('Best fit polarization angle:', round(polarization_angle.angle.degree, 3), '+/-', round(polarization_angle_uncertainty.degree, 3))
        
        return polarization



import numpy as np
import astropy.units as u
import functools
from abc import ABC, abstractmethod
import logging
logger = logging.getLogger(__name__)

class DeconvolutionAlgorithmBase(ABC):
    """
    A base class for image deconvolution algorithms.
    Subclasses should override these methods:

    - initialization 
    - pre_processing
    - Estep
    - Mstep
    - post_processing
    - register_result
    - check_stopping_criteria
    - finalization
    
    When the method run_deconvolution is called in ImageDeconvolution class, 
    the iteration method in this class is called for each iteration.

    Attributes
    ----------
    initial_model: :py:class:`cosipy.image_deconvolution.ModelBase` or its subclass
    dataset: list of :py:class:`cosipy.image_deconvolution.ImageDeconvolutionDataInterfaceBase` or its subclass
    parameter : py:class:`yayc.Configurator`
    results: list of results
    dict_bkg_norm: the dictionary of background normalizations
    dict_dataset_indexlist_for_bkg_models: the indices of data corresponding to each background model in the dataset
    """

    def __init__(self, initial_model, dataset, mask, parameter):

        self.initial_model = initial_model
        self.dataset = dataset
        self.mask = mask 
        self.parameter = parameter 
        self.results = []

        self.dict_bkg_norm = {}
        self.dict_dataset_indexlist_for_bkg_models = {}
        for data in self.dataset:
            for key in data.keys_bkg_models():
                if not key in self.dict_bkg_norm.keys():
                    self.dict_bkg_norm[key] = 1.0
                    self.dict_dataset_indexlist_for_bkg_models[key] = []
        
        for key in self.dict_dataset_indexlist_for_bkg_models.keys():
            for index, data in enumerate(self.dataset):
                if key in data.keys_bkg_models():
                    self.dict_dataset_indexlist_for_bkg_models[key].append(index)

        logger.debug(f'dict_bkg_norm: {self.dict_bkg_norm}')
        logger.debug(f'dict_dataset_indexlist_for_bkg_models: {self.dict_dataset_indexlist_for_bkg_models}')

        self.minimum_flux = parameter.get('minimum_flux:value', 0.0)

        minimum_flux_unit = parameter.get('minimum_flux:unit', initial_model.unit)
        if minimum_flux_unit is not None:
            self.minimum_flux = self.minimum_flux*u.Unit(minimum_flux_unit)

        # parameters of the iteration
        self.iteration_count = 0
        self.iteration_max = parameter.get('iteration_max', 1)

    @abstractmethod
    def initialization(self):
        """
        initialization before running the image deconvolution
        """
        raise NotImplementedError

    @abstractmethod
    def pre_processing(self):
        """
        pre-processing for each iteration
        """
        raise NotImplementedError

    @abstractmethod
    def Estep(self):
        """
        E-step. 
        In this step, only self.expectation_list should be updated.
        If it will be performed in other step, typically post_processing() or check_stopping_criteria(),
        this step can be skipped.
        """
        raise NotImplementedError

    @abstractmethod
    def Mstep(self):
        """
        M-step. 
        In this step, only self.delta_model should be updated.
        If you want to apply some operations to self.delta_model,
        these should be performed in post_processing().
        """
        raise NotImplementedError

    @abstractmethod
    def post_processing(self):
        """
        Post-processing for each iteration. 
        In this step, if needed, you can apply some filters to self.delta_model and set it as self.processed_delta_model.
        Then, the updated model should be calculated as self.model.
        For example, Gaussian smoothing can be applied to self.delta_model in this step.
        """
        raise NotImplementedError

    @abstractmethod
    def register_result(self):
        """
        Register results at the end of each iteration. 
        Users can define what kinds of values are stored in this method.
        """
        raise NotImplementedError

    @abstractmethod
    def check_stopping_criteria(self) -> bool:
        """
        Check if the iteration process should be continued or stopped.
        When it returns True, the iteration will stopped.
        """
        raise NotImplementedError

    @abstractmethod
    def finalization(self):
        """
        finalization after running the image deconvolution
        """
        raise NotImplementedError

### A subclass should not override the methods below. ###

    def iteration(self):
        """
        Perform one iteration of image deconvolution.
        This method should not be overrided in subclasses.
        """
        self.iteration_count += 1

        logger.info(f"## Iteration {self.iteration_count}/{self.iteration_max} ##")

        logger.info("<< Pre-processing >>")
        self.pre_processing()

        logger.info("<< E-step >>")
        self.Estep()

        logger.info("<< M-step >>")
        self.Mstep()
            
        logger.info("<< Post-processing >>")
        self.post_processing()

        logger.info("<< Registering Result >>")
        self.register_result()

        logger.info("<< Checking Stopping Criteria >>")
        stop_iteration = self.check_stopping_criteria()
        logger.info("--> {}".format("Stop" if stop_iteration else "Continue"))

        return stop_iteration

    def calc_expectation_list(self, model, dict_bkg_norm = None, almost_zero = 1e-12):
        """
        Calculate a list of expected count histograms corresponding to each data in the registered dataset.

        Parameters
        ----------
        model: :py:class:`cosipy.image_deconvolution.ModelBase` or its subclass
            Model
        dict_bkg_norm : dict, default None
            background normalization for each background model, e.g, {'albedo': 0.95, 'activation': 1.05}
        almost_zero : float, default 1e-12
            In order to avoid zero components in extended count histogram, a tiny offset is introduced.
            It should be small enough not to effect statistics.

        Returns
        -------
        list of :py:class:`histpy.Histogram`
            List of expected count histograms
        """
        
        return [data.calc_expectation(model, dict_bkg_norm = dict_bkg_norm, almost_zero = almost_zero) for data in self.dataset]

    def calc_loglikelihood_list(self, expectation_list):
        """
        Calculate a list of loglikelihood from each data in the registered dataset and the corresponding given expected count histogram.

        Parameters
        ----------
        expectation_list : list of :py:class:`histpy.Histogram`
            List of expected count histograms

        Returns
        -------
        list of float
            List of Log-likelood
        """

        return [data.calc_loglikelihood(expectation) for data, expectation in zip(self.dataset, expectation_list)]

    def calc_summed_exposure_map(self):
        """
        Calculate a list of exposure maps from the registered dataset.

        Returns
        -------
        list of :py:class:`histpy.Histogram`
        """

        return functools.reduce(lambda x, y: x + y, map(lambda data: data.exposure_map, self.dataset))

    def calc_summed_bkg_model(self, key):
        """
        Calculate the sum of histograms for a given background model in the registered dataset.

        Parameters
        ----------
        key: str
            Background model name

        Returns
        -------
        float
        """
        
        indexlist = self.dict_dataset_indexlist_for_bkg_models[key]
        
        return np.sum([self.dataset[i].summed_bkg_model(key) for i in indexlist])

    def calc_summed_T_product(self, dataspace_histogram_list):
        """
        For each data in the registered dataset, the product of the corresponding input histogram with the transonse of the response function is computed.
        Then, this method returns the sum of all of the products.

        Parameters
        ----------
        dataspace_histogram_list: list of :py:class:`histpy.Histogram`

        Returns
        -------
        :py:class:`histpy.Histogram`
        """

        return functools.reduce(lambda x, y: x + y, map(lambda data, hist: data.calc_T_product(hist), self.dataset, dataspace_histogram_list))

    def calc_summed_bkg_model_product(self, key, dataspace_histogram_list):
        """
        For each data in the registered dataset, the product of the corresponding input histogram with the specified background model is computed.
        Then, this method returns the sum of all of the products.

        Parameters
        ----------
        key: str
            Background model name
        dataspace_histogram_list: list of :py:class:`histpy.Histogram`

        Returns
        -------
        flaot
        """

        indexlist = self.dict_dataset_indexlist_for_bkg_models[key]

        return functools.reduce(lambda x, y: x + y, map(lambda i: self.dataset[i].calc_bkg_model_product(key = key, dataspace_histogram = dataspace_histogram_list[i]), indexlist))
